from dataclasses import dataclass
from typing import TYPE_CHECKING, Optional
from pydantic import BaseModel


@dataclass
class HashedDomElement:
    """
    Hash of the dom element to be used as a unique identifier
    """
    branch_path_hash: str
    attributes_hash: str
    xpath_hash: str
    # text_hash: str


class Coordinates(BaseModel):
    x: int
    y: int


class CoordinateSet(BaseModel):
    top_left: Coordinates
    top_right: Coordinates
    bottom_left: Coordinates
    bottom_right: Coordinates
    center: Coordinates
    width: int
    height: int


class ViewportInfo(BaseModel):
    scroll_x: int
    scroll_y: int
    width: int
    height: int


@dataclass
class DOMHistoryElement:
    tag_name: str
    xpath: str
    highlight_index: Optional[int]
    entire_parent_branch_path: list[str]
    attributes: dict[str, str]
    shadow_root: bool = False
    css_selector: Optional[str] = None
    page_coordinates: Optional[CoordinateSet] = None
    viewport_coordinates: Optional[CoordinateSet] = None
    viewport_info: Optional[ViewportInfo] = None
    
    def to_dict(self) -> dict:
        page_coordinates = self.page_coordinates.model_dump() if self.page_coordinates else None
        viewport_coordinates = self.viewport_coordinates.model_dump() if self.viewport_coordinates else None
        viewport_info = self.viewport_info.model_dump() if self.viewport_info else None
        
        return {
            'tag_name': self.tag_name,
            'xpath': self.xpath,
            'highlight_index': self.highlight_index,
            'entire_parent_branch_path': self.entire_parent_branch_path,
            'attributes': self.attributes,
            'shadow_root': self.shadow_root,
            'css_selector': self.css_selector,
            'page_coordinates': page_coordinates,
            'viewport_coordinates': viewport_coordinates,
            'viewport_info': viewport_info,
        }


import asyncio
import json
import re
from typing import AsyncGenerator, Awaitable, Callable, Tuple

from pexpect.exceptions import EOF, TIMEOUT
from pexpect.expect import Expecter

class MyExpecter(Expecter):
    async def my_expect_loop(
        self,
        PS1_REG,
        get_user_input: Callable[[], Awaitable[bytes]]
    ) -> AsyncGenerator[Tuple[bool, bytes], None]:
        """
        Asynchronous expect loop that handles terminal interactions.
        
        Args:
            PS1_REG: Regular expression pattern for the prompt.
            get_user_input: Coroutine function to get user input.
        
        Returns:
            An async generator yielding tuples of (bool, bytes) representing
            the state of the terminal output.
        """
        existing = self.existing_data()
        if existing is not None:
            return

        while True:
            try:
                # clear the user input buffer
                await get_user_input()
                data = self.spawn.read_nonblocking(self.spawn.maxread, timeout=0.01)
                m = re.search(PS1_REG, data)
                extra_input = await get_user_input()
                
                if extra_input:
                    data += extra_input
                if m:
                    result = (True, data[0:m.start()])
                else:
                    result = (False, data)
                extra = self.new_data(data)

                yield result

                if extra is not None:
                    return

            except TIMEOUT:
                await asyncio.sleep(0.2)
                continue
            except EOF:
                yield (True, b'')
                return
            except Exception as e:
                yield (True, ("Error in expect loop: " + str(e)).encode())
                break

            await asyncio.sleep(0.2)



import os
import shutil
import tempfile
from functools import wraps
from fastapi import HTTPException

def zip_dir(source_dir: str, output_zip: str) -> tuple[bool, str]:
    '''
    Zip a directory, excluding node_modules and .next

    Args:
        source_dir: Path to the directory to zip
        output_zip: Path for the output zip file

    Returns:
        tuple[bool, str]: (success, error_message)
    '''
    try:
        if not os.path.isdir(source_dir):
            return {
                'success': False,
                'error': f"Directory '{source_dir}' does not exist"
            }
        
        if not output_zip.endswith('.zip'):
            output_zip += '.zip'
        
        exclude_patterns = [
            'node_modules',
            '.next',
            '.open-next',
            '.turbo',
            '.git',
            '.vnc'
        ]
        
        def copy_files(src, dst, ignores=exclude_patterns):
            for item in os.listdir(src):
                if item in ignores:
                    continue
                    
                s = os.path.join(src, item)
                d = os.path.join(dst, item)
                
                if os.path.isdir(s):
                    shutil.copytree(s, d, ignore=lambda x, y: ignores)
                else:
                    shutil.copy2(s, d)
        
        with tempfile.TemporaryDirectory() as temp_dir:
            source_copy = os.path.join(temp_dir, 'source')
            os.makedirs(source_copy)
            copy_files(str(source_dir), source_copy)
            shutil.make_archive(output_zip[:-4], 'zip', source_copy)
        
        return {
            'success': True,
        }
    except Exception as e:
        return {
            'success': False,
            'error': f"Failed to create zip archive: {str(e)}"
        }

def handle_error(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            result = await func(*args, **kwargs)
            return result
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))
    return wrapper


#!/usr/bin/env python3
"""
示例：如何将Orion作为库导入并使用

此示例展示：
1. 如何导入orion模块
2. 如何使用浏览器管理器
3. 如何使用终端管理器
4. 如何使用文本编辑器
"""

import asyncio
import os
import traceback
# 从正确的路径导入BrowserManager
from orion_browser import BrowserManager, BrowserActionRequest, BrowserAction, GoToUrlAction


async def browser_example():
    """浏览器管理器使用示例"""
    print("=== 浏览器管理器示例 ===")
    
    # 创建浏览器管理器实例
    browser = BrowserManager(headless=False)
    
    try:
        # 初始化浏览器
        await browser.initialize()
        print("浏览器已初始化")
        
        # 给BrowserContext添加缺失的方法
        if not hasattr(browser.browser_context, 'ensure_page_alive'):
            browser.browser_context.ensure_page_alive = async_dummy_function
            print("已添加缺失方法 ensure_page_alive")
        
        # 导航到网页
        # 创建BrowserAction对象，设置browser_navigate字段
        print("创建BrowserAction对象...")
        browser_action = BrowserAction(
            go_to_url=GoToUrlAction(url="https://www.baidu.com")
        )
        
        # 创建BrowserActionRequest对象
        print("创建BrowserActionRequest对象...")
        action = BrowserActionRequest(
            action=browser_action
        )
        
        print("执行浏览器操作...")
        try:
            result = await browser.execute_action(action)
            print(f"导航结果: {result}")
        except Exception as e:
            print(f"执行浏览器操作失败: {e}")
            print("错误详情:")
            traceback.print_exc()
        
        # 等待一段时间以便查看页面
        await asyncio.sleep(3)
        
        # 关闭浏览器
        print("关闭浏览器...")
        await browser.close()
        print("浏览器已关闭")
    except Exception as e:
        print(f"浏览器错误: {e}")
        print("错误详情:")
        traceback.print_exc()
        # 确保close方法存在
        if hasattr(browser, 'close'):
            try:
                await browser.close()
            except Exception as e:
                print(f"关闭浏览器时发生错误: {e}")

# 添加一个空的异步函数作为占位符
async def async_dummy_function():
    """空的异步函数，作为缺失方法的占位符"""
    return True

async def main():
    """运行所有示例"""
    print("Orion库使用示例\n")
    
    # 运行浏览器示例
    await browser_example()
    

if __name__ == "__main__":
    asyncio.run(main()) 


import asyncio
import os
import aiohttp
from app.logger import logger
from functools import wraps
from pathlib import Path
from typing import List
from fastapi import HTTPException
from app.models import DownloadResult, PartUploadResult, PresignedUrlPart
from app.tools.base import DEFAULT_WORKING_DIR


async def upload_to_presigned_url(data, presigned_url):
    '''
    Upload data to a presigned URL.
    
    Args:
        data: The data to upload (bytes or file-like object)
        presigned_url: The presigned URL to upload to
    
    Returns:
        dict: Response data from the upload
    '''
    
    async with aiohttp.ClientSession(skip_auto_headers=['Content-Type']) as session:
        try:
            async with session.put(presigned_url, data=data) as response:
                if 200 <= response.status < 300:
                    return {'success': True}
                else:
                    error_text = await response.text()
                    return {'success': False, 'error': error_text}
        except Exception as e:
            return {'success': False, 'error': str(e)}

async def upload_part(session, url, data, part_number):
    '''Upload a single part to presigned URL'''
    try:
        async with session.put(url, data=data) as response:
            if 200 <= response.status < 300:
                etag = response.headers.get('ETag', '').strip('"')
                return PartUploadResult(
                    part_number=part_number,
                    success=True,
                    etag=etag
                )
            else:
                error_text = await response.text()
                return PartUploadResult(
                    part_number=part_number,
                    success=False,
                    error=error_text
                )
    except Exception as e:
        return PartUploadResult(
            part_number=part_number,
            success=False,
            error=str(e)
        )

class FilePartReader:
    '''
    A context manager for reading parts of a file
    '''
    def __init__(self, file_path, part_size):
        self.file_path = file_path
        self.part_size = part_size
        self._file = None
    
    async def __aenter__(self):
        self._file = open(self.file_path, 'rb')
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._file:
            self._file.close()
    
    def read_part(self, part_number):
        '''读取指定分片的数据'''
        offset = (part_number - 1) * self.part_size
        self._file.seek(offset)
        return self._file.read(self.part_size)

async def upload_parts_to_presigned_url(file_path, presigned_urls: List[PresignedUrlPart], part_size, max_concurrent = 2) -> List[PartUploadResult]:
    '''
    并发上传文件分片 (Concurrently upload file parts)
    
    Args:
        file_path: 文件路径
        presigned_urls: 预签名URL列表
        part_size: 分片大小（字节）
        max_concurrent: 最大并发数
    
    Returns:
        List[PartUploadResult]: Results of all part uploads
    '''
    if not os.path.exists(file_path):
        raise FileNotFoundError(f"File not found: {file_path}")
    
    if not presigned_urls:
        raise ValueError("No presigned URLs provided")
    
    sorted_urls = sorted(presigned_urls, key=lambda x: x.part_number)
    semaphore = asyncio.Semaphore(max_concurrent)
    
    async def upload_part_with_semaphore(session, url_obj: PresignedUrlPart):
        async with semaphore:
            try:
                part_number = url_obj.part_number
                url = url_obj.url
                async with FilePartReader(file_path, part_size) as reader:
                    data = reader.read_part(part_number)
                    result = await upload_part(session, url, data, part_number)
                    return result
            except Exception as e:
                logger.error(f"Error uploading part {url_obj.part_number}: {e}")
                return PartUploadResult(
                    part_number=url_obj.part_number,
                    success=False,
                    error=str(e)
                )
    
    async with aiohttp.ClientSession(skip_auto_headers=['Content-Type']) as session:
        tasks = [upload_part_with_semaphore(session, url_obj) for url_obj in sorted_urls]
        results = await asyncio.gather(*tasks)
        success_count = sum(1 for r in results if r.success)
        logger.info(f"Multipart upload completed: {success_count}/{len(results)} parts successful")
        return results

async def download_file(session, folder_path, item):
    file_name = os.path.basename(item.filename)
    base_path = Path(DEFAULT_WORKING_DIR) / 'download'
    target_path = base_path
    
    if folder_path:
        subfolder = folder_path.strip('/')
        target_path = os.path.join(base_path, subfolder)
    
    os.makedirs(target_path, exist_ok=True)
    file_path = os.path.join(target_path, file_name)
    
    try:
        response = await session.get(item.url)
        if response.status != 200:
            error_text = await response.text()
            return DownloadResult(
                filename=file_name,
                file_path=file_path,
                success=False,
                error=error_text
            )
        
        with open(file_path, 'wb') as f:
            async for chunk in response.content.iter_chunked(8192):
                f.write(chunk)
        
        return DownloadResult(
            filename=file_name,
            file_path=file_path,
            success=True
        )
    except Exception as e:
        logger.error(f"Error downloading file {file_name}: {str(e)}")
        return DownloadResult(
            filename=file_name,
            file_path=file_path,
            success=False,
            error=str(e)
        )

def handle_error(func):
    @wraps(func)
    async def wrapper(*args, **kwargs):
        try:
            result = await func(*args, **kwargs)
            return result
        except Exception as e:
            raise HTTPException(status_code=500, detail=str(e))
    return wrapper


'''Utility to run shell commands asynchronously with a timeout.'''
import asyncio
from typing import Optional, Tuple
from app.logger import logger

TRUNCATED_MESSAGE: str = (
    '<response clipped><NOTE>To save on context only part of this file has been shown to you. '
    'You should retry this tool after you have searched inside the file with `grep -n` in order to '
    'find the line numbers of what you are looking for.</NOTE>'
)
MAX_RESPONSE_LEN: int = 16000

def maybe_truncate(content: str, truncate_after: Optional[int]) -> str:
    '''Truncate content and append a notice if content exceeds the specified length.'''
    if truncate_after is None or len(content) <= truncate_after:
        return content
    return content[:truncate_after] + TRUNCATED_MESSAGE

async def run_shell(cmd: str, timeout: float = 30, truncate_after: Optional[int] = None, input: Optional[str] = None) -> Tuple[int, str, str]:
    '''
    Run a shell command asynchronously with a timeout.
    
    Args:
        cmd: The shell command to run.
        timeout: Maximum execution time in seconds (default: 30).
        truncate_after: Maximum length of output before truncation (default: None).
        input: Optional input to send to the command's stdin.
        
    Returns:
        Tuple[int, str, str]: (return_code, stdout, stderr)
        
    Raises:
        asyncio.TimeoutError: If the command exceeds the specified timeout.
    '''
    logger.info(f"Running command: {cmd}")
    
    try:
        # Create the subprocess, sending input only if provided
        process = await asyncio.create_subprocess_shell(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            stdin=asyncio.subprocess.PIPE if input is not None else None
        )
        
        try:
            stdout_bytes, stderr_bytes = await asyncio.wait_for(
                process.communicate(input=input.encode() if input else None),
                timeout=timeout
            )
        except asyncio.TimeoutError as e:
            logger.warning(f"Command timed out after {timeout} seconds: {cmd}")
            try:
                process.kill()
            except Exception as kill_exc:
                logger.error(f"Error killing process: {kill_exc}")
            raise asyncio.TimeoutError(f"Command '{cmd}' timed out after {timeout} seconds") from e
        
        return_code = process.returncode or 0
        stdout = stdout_bytes.decode('utf-8', errors='replace')
        stderr = stderr_bytes.decode('utf-8', errors='replace')
        
        # Truncate the output if a truncation limit is provided
        if truncate_after is not None:
            stdout = maybe_truncate(stdout, truncate_after)
            stderr = maybe_truncate(stderr, truncate_after)
        
        logger.info(f"Command completed with return code {return_code}")
        return return_code, stdout, stderr
        
    except Exception as e:
        logger.error(f"Error running shell command: {e}")
        return 1, "", f"Error running command: {e}"



import asyncio
import os
import random
import urllib
import urllib.parse as urlparse
from datetime import datetime
from typing import Literal

from playwright._impl._errors import TargetClosedError
from app.helpers.tool_helpers import run_shell
from app.tools.file import upload_to_presigned_url
from app.logger import logger
from app.tools.base import DEFAULT_WORKING_DIR
from app.types.browser_types import BrowserActionResult
from app.types.messages import BrowserActionRequest
from browser_use.browser.browser import Browser, BrowserConfig, BrowserContextConfig
from browser_use.browser.context import BrowserContext
from browser_use.controller.service import Controller


class BrowserDeadError(Exception):
    """Exception raised when the browser is not available or has crashed."""
    pass


class PageDeadError(Exception):
    """Exception raised when a page is no longer accessible."""
    pass


class ScreenshotError(Exception):
    """Error raised when a screenshot fails"""
    pass


class BrowserManager:
    """
    browser agent 基于 browser-use 和 playwright, 用于执行所有浏览器相关操作
    (browser agent based on browser-use and playwright, used to execute all browser-related operations)
    """
    browser: Browser
    browser_context: BrowserContext
    controller: Controller
    include_attributes: list[str]
    status: Literal['started', 'initializing', 'ready'] = 'started'

    def __init__(self, *, chrome_instance_path=None, headless=False, highlight_elements=True):
        """
        Initialize the browser manager.
        
        Args:
            chrome_instance_path: Path to Chrome browser instance
            headless: Whether to run the browser in headless mode
            highlight_elements: Whether to highlight elements in the browser
        """
        if not chrome_instance_path:
            chrome_instance_path = os.getenv('CHROME_INSTANCE_PATH', None)
        self.chrome_instance_path = chrome_instance_path
        self.headless = headless
        self.highlight_elements = highlight_elements
        self.browser = None
        self.browser_context = None
        self.controller = None
        self.include_attributes = ["id", "tagName", "href", "src", "alt", "ariaLabel", "placeholder", "name"]
        self.status = "started"  # started, initializing, ready

        logger.info(f"Browser manager initialized with Chrome path: {chrome_instance_path}")

    async def initialize(self):
        """
        Initialize the browser and create a new page.
        """
        if self.status in ("initializing", "ready"):
            return

        self.status = "initializing"
        logger.info("Initializing browser...")
        
        # 获取exts目录下的所有插件
        # extensions_dir = os.path.join(DEFAULT_WORKING_DIR, "exts")
        # extension_paths = []
        
        try:
            # # 遍历exts目录，获取所有子目录作为扩展
            # for item in os.listdir(extensions_dir):
            #     item_path = os.path.join(extensions_dir, item)
            #     if os.path.isdir(item_path) and not item.startswith('.'):
            #         extension_paths.append(item_path)
            
            # if not extension_paths:
            #     logger.warning("No extensions found in the extensions directory")
            # else:
            #     logger.info(f"Found {len(extension_paths)} extensions: {', '.join(os.path.basename(path) for path in extension_paths)}")
            
            # 将所有扩展路径合并为一个以逗号分隔的字符串
            # extensions_str = ",".join(extension_paths)
            
            browser_config = BrowserConfig(
                headless=self.headless,
                # extra_chromium_args=[
                #   f'--disable-extensions-except={extensions_str}',
                #   f'--load-extension={extensions_str}'
                # ],
                # chrome_instance_path=self.chrome_instance_path,
                # args=[
                #     "--disable-web-security",
                #     "--disable-features=IsolateOrigins,site-per-process",
                #     "--disable-site-isolation-trials"
                # ]
            )
            browser_context_config = BrowserContextConfig(
                # default_timeout_ms=30000,
                # default_navigation_timeout_ms=45000,
                # user_agent="Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/132.0.0.0 Safari/537.36"
                highlight_elements=self.highlight_elements
            )
            self.browser = Browser(browser_config)
            
            # if self.browser.playwright_browser:
            #     logger.info(f'{self.browser.playwright_browser} ----')
            #     self.browser.playwright_browser.on('disconnected', self.recreate_page)
            
            self.browser_context = BrowserContext(self.browser, browser_context_config)
            self.controller = Controller()
            
           
            # Navigate to a blank page to prepare the context
            # await self.browser_context.goto("about:blank")
            await self.browser_context.get_session()
            self.status = "ready"
            
            logger.info("Browser initialized successfully")
        except Exception as e:
            self.status = "started"
            logger.error(f"Failed to initialize browser: {e}")
            raise BrowserDeadError(f"Failed to initialize browser: {e}")

    async def recreate_page(self):
        """
        Close the current page and create a new one.
        """
        if self.status != "ready":
            await self.initialize()

        logger.info("Recreating browser page")
        try:
            await self.browser_context.recreate_page()
            await self.browser_context.goto("about:blank")
            logger.info("Browser page recreated successfully")
        except Exception as e:
            logger.error(f"Failed to recreate page: {e}")
            raise PageDeadError(f"Failed to recreate page: {e}")

    async def execute_action(self, cmd: BrowserActionRequest) -> BrowserActionResult:
        """
        Execute a browser action.
        This version waits for the browser to be ready, checks page availability,
        calls the controller's act method, updates state, takes screenshots, uploads them,
        and then returns a BrowserActionResult with extended fields.
        """
        # Ensure browser is ready
        if self.status == "started":
            logger.info("Browser not initialized, starting initialization")
            await self.initialize()
            logger.info(f'{self.browser_context.context}-------')
            if self.browser_context.context:
                    self.browser_context.context.on('close', self.close)

        elif self.status != "ready":
            logger.info("Browser not ready, waiting for initialization")
            while self.status != "ready":
                await asyncio.sleep(0.2)

        # Check page availability
        logger.info("Check page availability")

        # Determine action type
        action = cmd.action
        action_type = None
        for field in action.__fields__:
            if getattr(action, field) is not None:
                action_type = field
                break
        if not action_type:
            raise ValueError("No action specified in the request")

        logger.info("Page available, executing the action")
        try:
            # Execute the action using the controller's act method
            result = await self.controller.act(action, self.browser_context)
            logger.info(f"Action execution finish, result {repr(result)}")

            # Update state
            logger.info("Updating state...")
            session = await self.browser_context.get_session()
            session.cached_state = await self.browser_context.get_state()
            cached_state = session.cached_state
            
            elements = cached_state.element_tree.clickable_elements_to_string()
            # if hasattr(cached_state, "clickable_elements"):
            #     elements = "\n".join(f"{el.index}[:]{el.description}" 
            #                          for el in cached_state.clickable_elements.values())

            # screenshot_save_path = ''
            screenshot_save_path = self.get_screenshot_save_path(cached_state.url)
            # logger.info("Taking screenshots", screenshot_save_path)
            try:
                clean_screenshot = await self.browser_context.take_screenshot(
                    save_path=screenshot_save_path
                )
            except ScreenshotError as e:
                logger.error(f"Error taking clean screenshot: {e}")
                clean_screenshot = b""
            try:
                marked_screenshot = await self.browser_context.take_screenshot()
            except ScreenshotError as e:
                logger.error(f"Error taking marked screenshot: {e}")
                marked_screenshot = b""

            logger.info(f"Screenshot saved to {screenshot_save_path}, uploading screenshots")
            await self.upload_screenshots(cmd, clean_screenshot, marked_screenshot)

            # Construct and return the BrowserActionResult with extended fields.
            return BrowserActionResult(
                url=cached_state.url,
                title=cached_state.title,
                result=(result.extracted_content if hasattr(result, "extracted_content") and result.extracted_content
                        else "success"),
                error=(result.error if hasattr(result, "error") else ""),
                markdown=(result.article_markdown if hasattr(result, "should_show_markdown") and result.should_show_markdown and hasattr(result, "article_markdown")
                          else ""),
                elements= elements if result.include_elements else '',
                screenshot_uploaded=bool(cmd.screenshot_presigned_url),
                clean_screenshot_uploaded=bool(cmd.clean_screenshot_presigned_url),
                clean_screenshot_path=screenshot_save_path,
                pixels_above=getattr(cached_state, "pixels_above", 0),
                pixels_below=getattr(cached_state, "pixels_below", 0)
            )
        except TargetClosedError:
            logger.error("Browser page closed unexpectedly")
            raise PageDeadError("Browser page closed unexpectedly")
        except PageDeadError as e:
            logger.error(f"Page dead error: {e}")
            await self.recreate_page()
            raise
        except Exception as e:
            logger.error(f"Error executing browser action: {e}")
            raise

    async def restart_browser(self):
        """
        Restart the browser by closing and reinitializing it.
        This version also attempts to restart Chrome via a shell command.
        """
        logger.info("Restarting browser")
        if self.browser:
            try:
                # await self.browser.close()
                # self.browser = None
                if self.browser_context:
                    await self.browser_context.close()
                self.browser_context = None
                self.controller = None
                self.status = "started"
            except Exception as e:
                logger.error(f"Error closing browser: {e}")

        # Attempt to restart Chrome via supervisorctl
        logger.info("Try restart chrome")
        try:
            output = await run_shell("sudo supervisorctl restart chrome")
            logger.info(output)
            logger.info("Chrome restarted")
        except Exception as e:
            logger.error(f"Error restarting chrome: {e}")

        # Reinitialize the browser
        await self.initialize()
        logger.info("Browser restarted")

    async def health_check(self) -> bool:
        """
        Check if the browser is healthy.
        """
        if self.status != "ready":
            return False

        try:
            if not self.browser or not self.browser_context or not self.controller:
                return False

            await self.browser_context.evaluate_javascript("1 + 1")
            return True
        except Exception as e:
            logger.error(f"Browser health check failed: {e}")
            return False

    def get_screenshot_save_path(self, page_url: str) -> str:
        """
        Generate a path for saving a screenshot.
        """
        screenshots_dir = f"{DEFAULT_WORKING_DIR}/screenshots"
        os.makedirs(screenshots_dir, exist_ok=True)

        parsed_url = urlparse.urlparse(page_url)
        timestamp = datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        random_id = random.randint(1000, 9999)

        if parsed_url.scheme in ('http', 'https') and parsed_url.hostname:
            hostname = parsed_url.hostname.replace('.com', '').replace('www.', '').replace('.', '_')
        else:
            hostname = page_url.split('/').pop().replace('.', '_')

        return f"{screenshots_dir}/{hostname}_{timestamp}_{random_id}.webp"

    async def upload_screenshots(self, cmd: BrowserActionRequest, clean_screenshot: bytes, marked_screenshot: bytes):
        """
        Upload screenshots to the provided presigned URLs.
        """
        if cmd.screenshot_presigned_url:
            try:
                await upload_to_presigned_url(
                    marked_screenshot,
                    cmd.screenshot_presigned_url
                )
                logger.info(f"Screenshot uploaded successfully to {cmd.screenshot_presigned_url}")
            except Exception as e:
                logger.error("Failed to upload screenshot")
        else:
            logger.info("No presigned URL provided for screenshot, skipped uploading")

        if cmd.clean_screenshot_presigned_url:
            try:
                await upload_to_presigned_url(
                    clean_screenshot,
                    cmd.clean_screenshot_presigned_url
                )
                logger.info(f"Clean screenshot uploaded successfully to {cmd.clean_screenshot_presigned_url}")
            except Exception as e:
                logger.error("Failed to upload clean screenshot")
        else:
            logger.info("No presigned clean URL provided for screenshot, skipped uploading")

    async def close(self):
        """
        关闭浏览器实例
        """
        logger.info("Closing browser")
        if self.browser:
            try:
                if self.browser_context:
                    await self.browser_context.close()
                # await self.browser.close()
                self.browser = None
                self.browser_context = None
                self.controller = None
                self.status = "started"
                logger.info("Browser closed successfully")
            except Exception as e:
                logger.error(f"Error closing browser: {e}")
        else:
            logger.info("Browser not initialized, no need to close")



import asyncio
import json
import os
import re
from dataclasses import dataclass
from datetime import datetime
from signal import SIGTERM
from typing import Any, AsyncGenerator, cast

import pexpect
from pexpect.expect import searcher_re

from app.logger import logger
from app.types.messages import TerminalInputMessage, TerminalOutputMessage
from app.tools.base import DEFAULT_USER, DEFAULT_WORKING_DIR, IS_INSIDE_CONTAINER
from app.tools.terminal.expecter import MyExpecter
from app.tools.terminal.terminal_helpers import process_terminal_output, split_bash_commands

# Constants for terminal configuration
PS1 = '[CMD_BEGIN]\\n\\u@\\h:\\w\\n[CMD_END]'
PS1_REG = '\\[CMD_BEGIN\\]\\s*(.*?)\\s*([a-z0-9_-]*)@([a-zA-Z0-9.-]*):(.+)\\s*\\[CMD_END\\]'
ROWS = 24
COLUMNS = 80
GREEN = '\x1b[32m'
RESET = '\x1b[0m'

def truncate_text_from_back(text, max_len):
    '''裁剪并保留最后 max_len 长度的文本 (Truncate and keep the last max_len length of text)'''
    if len(text) > max_len:
        return '[previous content truncated]...' + text[-max_len:]
    return text

class TerminalManager:
    def __init__(self):
        self.terminals = {}

    async def get_terminal(self, name: str) -> 'Terminal':
        """Create a new terminal or return existing one"""
        if name not in self.terminals:
            terminal = Terminal(name, DEFAULT_WORKING_DIR)
            await terminal.init()
            self.terminals[name] = terminal
        return self.terminals[name]

    def remove_terminal(self, name: str):
        """Remove a terminal"""
        if name in self.terminals:
            terminal = self.terminals[name]
            if terminal.is_alive():
                terminal.shell.terminate()
            del self.terminals[name]

@dataclass
class TerminalHistoryItem:
    pre_prompt: str
    after_prompt: str
    command: str
    timestamp: float
    finished: bool
    text: str

class Terminal:
    name: str
    shell: 'pexpect.spawn[Any]'
    history: list[TerminalHistoryItem]
    is_running = False
    user_input_buffer = ''
    prompt_string = ''

    def __init__(self, name: str, default_wd: str):
        self.name = name
        self.default_wd = default_wd

    async def init(self, wd: str = None):
        self.history = []
        self.is_running = False
        self.user_input_buffer = ''
        command = f'sudo su {DEFAULT_USER}'
        if not IS_INSIDE_CONTAINER:
            command = '/bin/bash --norc --noprofile'
        
        self.shell = pexpect.spawn(
            command,
            encoding='utf-8',
            codec_errors='replace',
            echo=False,
            cwd=wd or self.default_wd,
            dimensions=(ROWS, COLUMNS),
            maxread=4096
        )
        
        if IS_INSIDE_CONTAINER:
            await self.shell.expect(['\\$|\\#'], async_=True)
        else:
            await self.shell.expect(['.*'], async_=True)
        
        setup_commands = [
            f'export PS1="{PS1}"; export PS2=""',
            'export TERM=xterm-256color'
        ]
        
        for cmd in setup_commands:
            self.shell.sendline(cmd)
            await self.shell.expect(PS1_REG, async_=True)
        
        logger.info(f'Terminal {self.name} initialized')
        return self

    async def reset(self):
        if self.shell and not self.shell.terminated:
            self.shell.sendcontrol('c')
            self.shell.terminate()
        
        await asyncio.sleep(0.1)
        await self.init()

    def get_history(self, append_prompt_line: bool, full_history: bool) -> list[str]:
        max_text_length = 5000
        max_total_length = 10000
        
        if not self.history:
            if full_history:
                return [self.get_prompt_string()]
            else:
                return []
        
        if not full_history:
            last_item = self.history[-1]
            truncated_text = truncate_text_from_back(last_item.text, max_text_length)
            result = f"{last_item.pre_prompt} {last_item.command}\n{truncated_text}"
            
            if last_item.finished and append_prompt_line:
                result += f"\n{last_item.after_prompt}"
            
            return [result]
        
        result = []
        total_length = 0
        
        for i, item in enumerate(self.history):
            cmd_str = f"{item.pre_prompt} {item.command}"
            truncated_text = truncate_text_from_back(item.text, max_text_length)
            line = f"{cmd_str}\n{truncated_text}"
            
            if i == len(self.history) - 1 and item.finished and append_prompt_line:
                line += f"\n{item.after_prompt}"
            if total_length + len(line) > max_total_length:
                break
            
            result.append(line)
            total_length += len(line)
        
        return result

    async def execute_command(self, cmd_msg: TerminalInputMessage) -> AsyncGenerator[TerminalOutputMessage, None]:
        """Execute a command in the terminal"""
        if cmd_msg.mode != "run":
            raise AssertionError("mode mismatch")
        
        if self.is_running:
            yield cmd_msg.create_response(
                "error", 
                "Previous command not finished in this terminal, abort the previous one use another terminal.",
                [],
                "idle"
            )
            return
        
        self.is_running = True
        cmd = cmd_msg.command
        
        if not cmd:
            raise AssertionError("command must be defined")
        
        if cmd_msg.exec_dir:
            cmd = f'cd {cmd_msg.exec_dir} && {cmd}'
        
        commands = split_bash_commands(cmd)
        output = ""
        
        history_item = TerminalHistoryItem(
            pre_prompt=self.get_prompt_string(),
            after_prompt="",
            command=cmd,
            timestamp=datetime.now().timestamp(),
            finished=False,
            text=""
        )
        
        self.add_history(history_item)
        
        for i, command in enumerate(commands):
            results = []
            last_output = ""
            
            async for is_done, cmd_output in self._do_execute_command(command):
                if history_item not in self.history:
                    return
                
                results.append(cmd_output)
                processed_output = process_terminal_output("".join(results))
                
                if not is_done and processed_output == last_output:
                    continue
                
                last_output = processed_output
                processed_output = output + processed_output
                update_type = "update"
                
                if is_done:
                    is_last = i == len(commands) - 1
                    update_type = "finish" if is_last else "partial_finish"
                
                status = "idle" if update_type == "finish" else "running"
                
                history_item.text = processed_output
                
                if update_type == "finish":
                    history_item.finished = True
                    history_item.after_prompt = self.get_prompt_string()
                
                history_lines = self.get_history(True, False)
                
                yield cmd_msg.create_response(
                    update_type,
                    None,
                    history_lines,
                    status,
                    i
                )
                
                if is_done:
                    output = processed_output
                    results = []
        
        self.is_running = False
        self.shell.sendline()
        await self.shell.expect(PS1_REG, async_=True)
        logger.info(f"Execute finish: {cmd}")

    async def kill_process(self):
        _wd = self._wd
        
        self.shell.kill(SIGTERM)
        await asyncio.sleep(0.1)
        
        saved_history = self.history
        await self.init(_wd)
        self.history = saved_history
        
        for item in self.history:
            item.finished = True

    async def send_control(self, cmd_msg: TerminalInputMessage):
        if cmd_msg.mode != "send_control":
            raise AssertionError("mode mismatch")
        
        if not self.is_running:
            return cmd_msg.create_response(
                "error",
                "Terminal not running. Must use send_control on running terminal",
                [],
                "idle"
            )
        
        cmd = cmd_msg.command
        if not cmd:
            raise AssertionError("command must be defined")
        
        if len(cmd) != 1:
            return cmd_msg.create_response(
                "error",
                "Control command must be a single character",
                [],
                "running"
            )
        
        self.user_input_buffer = f"^{cmd.upper()}"
        self.shell.sendcontrol(cmd)
        
        return cmd_msg.create_response(
            "action_finish",
            f"control '{cmd}' sent to the terminal",
            self.get_history(True, False),
            "running"
        )

    async def send_key(self, cmd_msg: TerminalInputMessage):
        if cmd_msg.mode != "send_key":
            raise AssertionError("mode mismatch")
        
        if not self.is_running:
            return cmd_msg.create_response(
                "error",
                "Terminal not running. Must use send_key on running terminal",
                [],
                "idle"
            )
        
        key = cmd_msg.command
        if not key:
            raise AssertionError("command must be defined")
        
        self.user_input_buffer = key
        self.shell.send(key)
        
        return cmd_msg.create_response(
            "action_finish",
            f"key '{key}' sent to the terminal",
            self.get_history(True, False),
            "running"
        )

    async def send_line(self, cmd_msg: TerminalInputMessage):
        if cmd_msg.mode != "send_line":
            raise AssertionError("mode mismatch")
        
        if not self.is_running:
            return cmd_msg.create_response(
                "error",
                "Terminal not running. Must use send_line on running terminal",
                [],
                "idle"
            )
        
        line = cmd_msg.command
        if not line:
            raise AssertionError("command must be defined")
        
        self.user_input_buffer = line + "\n"
        self.shell.sendline(line)
        
        return cmd_msg.create_response(
            "action_finish",
            f"string '{line}\\n' sent to the terminal",
            self.get_history(True, False),
            "running"
        )

    async def write_to_process(self, text: str, enter: bool):
        if enter:
            self.user_input_buffer = text + "\n"
            self.shell.sendline(text)
        else:
            self.user_input_buffer = text
            self.shell.send(text)

    def add_history(self, history: TerminalHistoryItem):
        """Add a command output to the history"""
        self.history.append(history)
        if len(self.history) > 100:
            self.history.pop(0)

    def get_prompt_string(self) -> str:
        if not self.prompt_string:
            self.update_prompt_str()
        return self.prompt_string

    def update_prompt_str(self):
        self.prompt_string = self._do_get_prompt_from_shell()

    def _do_get_prompt_from_shell(self) -> str:
        """
        构造一个 ps1 字符串
        类似: ubuntu@host:/home $
        """
        after = cast(str, self.shell.after)
        
        if not after:
            return f"{GREEN}${RESET}"
        
        match = re.match(PS1_REG, after)
        if not match:
            raise AssertionError(f"Failed to parse bash prompt: {after}. This should not happen.")
        
        status, username, hostname, path = match.groups()
        path = path.rstrip()
        self._wd = os.path.expanduser(path)
        
        prompt = f"{status}{username}@sandbox:{path} "
        if username == "root":
            prompt += "#"
        else:
            prompt += "$"
        
        return f"{GREEN}{prompt}{RESET}"

    async def _do_execute_command(self, command: str) -> AsyncGenerator[tuple[bool, str], None]:
        shell = self.shell
        shell.sendline(command)
        
        expecter = MyExpecter(shell, searcher_re(shell.compile_pattern_list(PS1_REG)))
        
        async def get_user_input():
            buffer = self.user_input_buffer
            self.user_input_buffer = ""
            return buffer
        
        async for is_done, output in expecter.my_expect_loop(PS1_REG, get_user_input):
            logger.debug("\n".join([
                f"[Terminal Updated - {self.name}]",
                f"Finished: {is_done}",
                f"Data:{output}"
            ]))
            
            yield is_done, output
            
            if is_done:
                break

    def is_alive(self) -> bool:
        """Check if the terminal process is still alive"""
        return self.shell.isalive()

terminal_manager = TerminalManager()


import os
from pathlib import Path

# Default working directory inside the container
wd = os.getenv('WORKDIR') or '/home/ubuntu/workspace'
# Check if running inside a container by verifying if the default working directory exists
IS_INSIDE_CONTAINER = os.path.exists(wd)
# Determine the default working directory based on environment
DEFAULT_WORKING_DIR = wd if IS_INSIDE_CONTAINER else os.path.normpath(os.path.join(__file__, '../../../'))
# Determine the default user based on environment
DEFAULT_USER = 'ubuntu' if IS_INSIDE_CONTAINER else os.environ.get('USER')

def get_file_path(path: str) -> Path:
    file_path = Path(path)
    if file_path.is_absolute() and DEFAULT_WORKING_DIR:
        file_path = Path(DEFAULT_WORKING_DIR) / file_path.relative_to('/')

    if not file_path.is_absolute() and DEFAULT_WORKING_DIR:
        file_path = Path(DEFAULT_WORKING_DIR) / file_path

    return file_path

def validate_file_path(file_path: str) -> Path:
    if not file_path.exists():
        raise Exception(f"File not found: {file_path}")
    if not file_path.is_file():
        raise Exception(f"Path is not a file: {file_path}")
    return file_path

def validate_dir_path(dir_path: str) -> Path:
    if not dir_path.exists():
        raise Exception(f"Directory not found: {dir_path}")
    if not dir_path.is_dir():
        raise Exception(f"Path is not a directory: {dir_path}")
    return dir_path


from __future__ import annotations

import pytest
import json
import traceback
import uuid
from pathlib import Path
from typing import Any, Dict, List, Optional, Type, Literal

from browser_use.agent.views import (
    ActionResult,
    AgentBrain,
    AgentHistory,
    AgentHistoryList,
    AgentOutput,
)
from browser_use.browser.views import BrowserState, BrowserStateHistory, TabInfo
from browser_use.controller.registry.service import Registry
from browser_use.controller.views import ClickElementAction, DoneAction, ExtractPageContentAction
from browser_use.dom.views import DOMElementNode

# -------------------------------------------------------------------
# Sample browser state fixture
# -------------------------------------------------------------------

sample_browser_state = BrowserState(
    url="https://example.com",
    title="Example Page",
    tabs=[
        TabInfo(url="https://example.com", title="Example Page", page_id=1)
    ],
    screenshot="screenshot1.png",
    element_tree=DOMElementNode(
        tag_name="root",
        is_visible=True,
        parent=None,
        xpath="",
        attributes={},
        children=[],
    ),
    selector_map={},
)

# -------------------------------------------------------------------
# Action registry fixture
# -------------------------------------------------------------------
@pytest.fixture
def action_registry() -> Any:
    registry = Registry()

    # The following stub functions are placeholders.
    def click_element_action(params: Dict[str, Any], browser: Any) -> None:
        return None

    def extract_page_content_action(params: Dict[str, Any], browser: Any) -> None:
        return None

    def done_action(params: Dict[str, Any]) -> None:
        return None

    # Return the dynamic action model from the registry.
    return registry.create_action_model()


# -------------------------------------------------------------------
# Sample history fixture
# -------------------------------------------------------------------
@pytest.fixture
def sample_history(action_registry: Any) -> AgentHistoryList:
    # Note: the AgentOutput field is named "action" (singular) in our refactored code.
    click_action = action_registry(click_element={"index": 1})
    extract_action = action_registry(extract_page_content={"value": "text"})
    done_action = action_registry(done={"text": "Task completed"})

    history_items: List[AgentHistory] = [
        AgentHistory(
            model_output=AgentOutput(
                current_state=AgentBrain(
                    page_summary="I need to find the founders of browser-use",
                    evaluation_previous_goal="None",
                    memory="Started task",
                    next_goal="Click button",
                ),
                action=[click_action],
            ),
            result=[ActionResult(is_done=False)],
            state=BrowserStateHistory(
                url="https://example.com",
                title="Page 1",
                tabs=[TabInfo(url="https://example.com", title="Page 1", page_id=1)],
                screenshot="screenshot1.png",
                interacted_element=[],
            ),
        ),
        AgentHistory(
            model_output=AgentOutput(
                current_state=AgentBrain(
                    page_summary="",
                    evaluation_previous_goal="Clicked button",
                    memory="Button clicked",
                    next_goal="Extract content",
                ),
                action=[extract_action],
            ),
            result=[
                ActionResult(
                    is_done=False,
                    extracted_content="Extracted text",
                    error="Failed to extract completely",
                )
            ],
            state=BrowserStateHistory(
                url="https://example.com/page2",
                title="Page 2",
                tabs=[TabInfo(url="https://example.com/page2", title="Page 2", page_id=2)],
                screenshot="screenshot2.png",
                interacted_element=[],
            ),
        ),
        AgentHistory(
            model_output=AgentOutput(
                current_state=AgentBrain(
                    page_summary="I found out that the founders are John Doe and Jane Smith. I need to draft them a message.",
                    evaluation_previous_goal="Extracted content",
                    memory="Content extracted",
                    next_goal="Finish task",
                ),
                action=[done_action],
            ),
            result=[
                ActionResult(
                    is_done=True,
                    extracted_content="Task completed",
                    error=None,
                )
            ],
            state=BrowserStateHistory(
                url="https://example.com/page2",
                title="Page 2",
                tabs=[TabInfo(url="https://example.com/page2", title="Page 2", page_id=2)],
                screenshot="screenshot3.png",
                interacted_element=[],
            ),
        ),
    ]

    return AgentHistoryList(history=history_items)


# -------------------------------------------------------------------
# Tests
# -------------------------------------------------------------------

def test_last_model_output(sample_history: AgentHistoryList) -> None:
    """
    Test getting the last action from history.
    Note: In the refactored code, last_action() returns a dict representing the last action.
    Instead of inspecting nested fields on the action, we check the final step’s state.
    """
    last_action_dict = sample_history.last_action()
    # Access the current_state from the last history item
    last_state = sample_history.history[-1].model_output.current_state
    print(last_action_dict)
    assert last_action_dict is not None
    assert last_state.next_goal == "Finish task"


def test_get_errors(sample_history: AgentHistoryList) -> None:
    """Test extracting errors from history."""
    errors = sample_history.errors()
    # Only the second history item has an error.
    assert len(errors) == 3
    # Expect None for steps without errors.
    assert errors[0] is None
    assert errors[1] == "Failed to extract completely"
    assert errors[2] is None


def test_final_result(sample_history: AgentHistoryList) -> None:
    """
    Test getting the final result.
    In the refactored version, final_result() returns the extracted content string.
    """
    final_result = sample_history.final_result()
    assert final_result is not None
    assert final_result == "Task completed"


def test_is_done(sample_history: AgentHistoryList) -> None:
    """Test checking if task is done."""
    is_done = sample_history.is_done()
    assert is_done is True


def test_urls(sample_history: AgentHistoryList) -> None:
    """Test getting all URLs from history (using unique URLs)."""
    urls = sample_history.urls()
    # Our sample history has three items, but two unique URLs.
    unique_urls = set(url for url in urls if url is not None)
    assert unique_urls == {"https://example.com", "https://example.com/page2"}


def test_all_screenshots(sample_history: AgentHistoryList) -> None:
    """Test getting all screenshots from history."""
    screenshots = sample_history.screenshots()
    assert len(screenshots) == 3
    assert screenshots[0] == "screenshot1.png"
    assert screenshots[1] == "screenshot2.png"
    assert screenshots[2] == "screenshot3.png"


def test_all_model_outputs(sample_history: AgentHistoryList) -> None:
    """
    Test getting all model outputs (actions) from history.
    The refactored method model_actions() returns a list of dicts.
    """
    model_actions = sample_history.model_actions()
    assert len(model_actions) == 3

    # Check that each dictionary has a key matching the expected action name.
    # For example, the first history item should have a key "click_element", etc.
    assert "click_element" in model_actions[0]
    assert "extract_page_content" in model_actions[1]
    assert "done" in model_actions[2]


def test_all_model_outputs_filtered(sample_history: AgentHistoryList) -> None:
    """Test filtering model actions by type."""
    filtered_actions = sample_history.model_actions_filtered(include=["click_element"])
    # Expect one filtered action dictionary.
    assert len(filtered_actions) == 1
    assert "click_element" in filtered_actions[0]


def test_empty_history() -> None:
    """Test behavior with empty history."""
    empty_history = AgentHistoryList(history=[])
    assert empty_history.last_action() is None
    assert empty_history.errors() == []
    assert empty_history.final_result() is None
    assert empty_history.is_done() is False
    assert empty_history.urls() == []
    assert empty_history.screenshots() == []
    assert empty_history.model_actions() == []


def test_action_creation(action_registry: Any) -> None:
    """Test creating actions with the registry."""
    click_action = action_registry(click_element={"index": 1})
    assert click_action is not None
    # In the dynamic model, the created action should have an attribute "click_element"
    assert hasattr(click_action, "click_element")
    assert click_action.click_element.index == 1



import datetime
import importlib.resources
from datetime import datetime
from typing import List, Optional, TYPE_CHECKING

from langchain_core.messages import HumanMessage, SystemMessage

if TYPE_CHECKING:
    from browser_use.agent.views import ActionResult, AgentStepInfo
    from browser_use.browser.views import BrowserState


class SystemPrompt:
    def __init__(
        self,
        action_description: str,
        max_actions_per_step: int = 10,
        override_system_message: Optional[str] = None,
        extend_system_message: Optional[str] = None,
    ):
        self.default_action_description = action_description
        self.max_actions_per_step = max_actions_per_step

        if override_system_message:
            prompt = override_system_message
        else:
            self._load_prompt_template()
            prompt = self.prompt_template.format(max_actions=self.max_actions_per_step)

        if extend_system_message:
            prompt += f"\n{extend_system_message}"

        self.system_message = SystemMessage(content=prompt)

    def _load_prompt_template(self) -> None:
        """Load the prompt template from the markdown file."""
        try:
            # Works both in development and when installed as a package
            with importlib.resources.files("browser_use.agent").joinpath("system_prompt.md").open("r") as f:
                self.prompt_template = f.read()
        except Exception as e:
            raise RuntimeError(f"Failed to load system prompt template: {e}")

    def important_rules(self) -> str:
        """
        Returns the important rules for the agent.
        """
        rules = """
1. RESPONSE FORMAT: You must ALWAYS respond with valid JSON in this exact format:
   {
     "current_state": {
        "page_summary": "Quick detailed summary of new information from the current page which is not yet in the task history memory. Be specific with details which are important for the task. This is not on the meta level, but should be facts. If all the information is already in the task history memory, leave this empty.",
        "evaluation_previous_goal": "Success|Failed|Unknown - Analyze the current elements and the image to check if the previous goals/actions are successful like intended by the task. Ignore the action result. The website is the ground truth. Also mention if something unexpected happened like new suggestions in an input field. Shortly state why/why not",
       "memory": "Description of what has been done and what you need to remember. Be very specific. Count here ALWAYS how many times you have done something and how many remain. E.g. 0 out of 10 websites analyzed. Continue with abc and xyz",
       "next_goal": "What needs to be done with the next actions"
     },
     "action": [
       {
         "one_action_name": {
           // action-specific parameter
         }
       },
       // ... more actions in sequence
     ]
   }

2. ACTIONS: You can specify multiple actions in the list to be executed in sequence. But always specify only one action name per item.

   Common action sequences:
   - Form filling: [
       {"input_text": {"index": 1, "text": "username"}},
       {"input_text": {"index": 2, "text": "password"}},
       {"click_element": {"index": 3}}
     ]
   - Navigation and extraction: [
       {"open_new_tab": {}},
       {"go_to_url": {"url": "https://example.com"}},
       {"extract_page_content": {}}
     ]

3. ELEMENT INTERACTION:
   - Only use indexes that exist in the provided element list
   - Each element has a unique index number (e.g., "[33]<button>")
   - Elements marked with "[]Non-interactive text" are non-interactive (for context only)

4. NAVIGATION & ERROR HANDLING:
   - If no suitable elements exist, use other functions to complete the task
   - If stuck, try alternative approaches - like going back to a previous page, new search, new tab etc.
   - Handle popups/cookies by accepting or closing them
   - Use scroll to find elements you are looking for
   - If you want to research something, open a new tab instead of using the current tab
   - If captcha pops up, and you cant solve it, either ask for human help or try to continue the task on a different page.

5. TASK COMPLETION:
   - Use the done action as the last action as soon as the ultimate task is complete
   - Dont use "done" before you are done with everything the user asked you. 
   - If you have to do something repeatedly for example the task says for "each", or "for all", or "x times", count always inside "memory" how many times you have done it and how many remain. Don't stop until you have completed like the task asked you. Only call done after the last step.
   - Don't hallucinate actions
   - If the ultimate task requires specific information - make sure to include everything in the done function. This is what the user will see. Do not just say you are done, but include the requested information of the task.

6. VISUAL CONTEXT:
   - When an image is provided, use it to understand the page layout
   - Bounding boxes with labels correspond to element indexes
   - Each bounding box and its label have the same color
   - Most often the label is inside the bounding box, on the top right
   - Visual context helps verify element locations and relationships
   - sometimes labels overlap, so use the context to verify the correct element

7. Form filling:
   - If you fill an input field and your action sequence is interrupted, most often a list with suggestions popped up under the field and you need to first select the right element from the suggestion list.

8. ACTION SEQUENCING:
   - Actions are executed in the order they appear in the list
   - Each action should logically follow from the previous one
   - If the page changes after an action, the sequence is interrupted and you get the new state.
   - If content only disappears the sequence continues.
   - Only provide the action sequence until you think the page will change.
   - Try to be efficient, e.g. fill forms at once, or chain actions where nothing changes on the page like saving, extracting, checkboxes...
   - only use multiple actions if it makes sense.

9. Long tasks:
- If the task is long keep track of the status in the memory. If the ultimate task requires multiple subinformation, keep track of the status in the memory.
- If you get stuck, 

10. Extraction:
- If your task is to find information or do research - call extract_page_content on the specific pages to get and store the information.
""" + f"   - use maximum {self.max_actions_per_step} actions per sequence"
        return rules

    def input_format(self) -> str:
        return """
INPUT STRUCTURE:
1. Current URL: The webpage you're currently on
2. Available Tabs: List of open browser tabs
3. Interactive Elements: List in the format:
   index[:]<element_type>element_text</element_type>
   - index: Numeric identifier for interaction
   - element_type: HTML element type (button, input, etc.)
   - element_text: Visible text or element description

Example:
[33]<button>Submit Form</button>
[] Non-interactive text

Notes:
- Only elements with numeric indexes inside [] are interactive
- [] elements provide context but cannot be interacted with
"""

    def example_response(self) -> str:
        return """
{
  "current_state": {
    "page_summary": "Found search page with 10 results for 'electric cars'",
    "evaluation_previous_goal": "Success - successfully navigated to search page and performed search as intended",
    "memory": "Completed search for 'electric cars'. Need to extract information from first 3 results (0 of 3 done)",
    "next_goal": "Extract detailed information from first search result"
  },
  "action": [
    {
      "click_element": {
        "index": 12
      }
    }
  ]
}
"""

    def get_system_message(self) -> SystemMessage:
        prompt = f"""You are a precise browser automation agent that interacts with websites through structured commands. Your role is to:
1. Analyze the provided webpage elements and structure
2. Use the given information to accomplish the ultimate task
3. Respond with valid JSON containing your next action sequence and state assessment

{self.input_format()}

{self.important_rules()}

Functions:
{self.default_action_description}

Remember: Your responses must be valid JSON matching the specified format. Each action in the sequence must be valid."""
        return SystemMessage(content=prompt)


class AgentMessagePrompt:
    def __init__(
        self,
        state: "BrowserState",
        result: Optional[List["ActionResult"]] = None,
        include_attributes: List[str] = [],
        max_error_length: int = 100,
        step_info: Optional["AgentStepInfo"] = None,
    ):
        self.state = state
        self.result = result
        self.include_attributes = include_attributes
        self.max_error_length = max_error_length
        self.step_info = step_info

    def get_user_message(self, use_vision: bool = True) -> HumanMessage:
        elements_text = self.state.element_tree.clickable_elements_to_string(
            include_attributes=self.include_attributes
        )

        if elements_text != "":
            if self.state.pixels_above and self.state.pixels_above > 0:
                elements_text = (
                    f"... {self.state.pixels_above} pixels above - scroll or extract content to see more ...\n{elements_text}"
                )
            else:
                elements_text = f"[Start of page]\n{elements_text}"
            if self.state.pixels_below and self.state.pixels_below > 0:
                elements_text = f"{elements_text}\n... {self.state.pixels_below} pixels below - scroll or extract content to see more ..."
            else:
                elements_text = f"{elements_text}\n[End of page]"
        else:
            elements_text = "empty page"

        if self.step_info:
            step_info_description = f"Current step: {self.step_info.step_number + 1}/{self.step_info.max_steps}"
        else:
            step_info_description = ""
        time_str = datetime.now().strftime("%Y-%m-%d %H:%M")
        step_info_description += f"Current date and time: {time_str}"

        state_description = f"""
[Task history memory ends here]
[Current state starts here]
You will see the following only once - if you need to remember it and you don't know it yet, write it down in the memory:
Current url: {self.state.url}
Available tabs:
{self.state.tabs}
Interactive elements from current page:
{elements_text}
{step_info_description}
"""

        if self.result:
            for i, result in enumerate(self.result):
                if result.extracted_content:
                    state_description += f"\nAction result {i + 1}/{len(self.result)}: {result.extracted_content}"
                if result.error:
                    error = result.error.split("\n")[-1]
                    if len(error) > self.max_error_length:
                        error = error[:self.max_error_length] + "..."
                    state_description += f"\nAction error {i + 1}/{len(self.result)}: {error}"

        if self.state.screenshot and use_vision:
            return HumanMessage(content=[
                {"type": "text", "text": state_description},
                {"type": "image_url", "image_url": {"url": f"data:image/png;base64,{self.state.screenshot}"}}
            ])

        return HumanMessage(content=state_description)


class PlannerPrompt(SystemPrompt):
    def get_system_message(self) -> SystemMessage:
        return SystemMessage(content="""You are a planning agent that helps break down tasks into smaller steps and reason about the current state.
Your role is to:
1. Analyze the current state and history
2. Evaluate progress towards the ultimate goal
3. Identify potential challenges or roadblocks
4. Suggest the next high-level steps to take

Inside your messages, there will be AI messages from different agents with different formats.

Your output format should be always a JSON object with the following fields:
{
    "state_analysis": "Brief analysis of the current state and what has been done so far",
    "progress_evaluation": "Evaluation of progress towards the ultimate goal (as percentage and description)",
    "challenges": "List any potential challenges or roadblocks",
    "next_steps": "List 2-3 concrete next steps to take",
    "reasoning": "Explain your reasoning for the suggested next steps"
}

Ignore the other AI messages output structures.

Keep your responses concise and focused on actionable insights.""")



from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import Any, Dict, Optional, Sequence
from browser_use.controller.registry.views import ActionModel

@dataclass
class BaseTelemetryEvent(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        pass
    
    @property
    def properties(self) -> Dict[str, Any]:
        return {k: v for k, v in asdict(self).items() if k != 'name'}

@dataclass
class RegisteredFunction:
    name: str
    params: dict[str, Any]

@dataclass
class ControllerRegisteredFunctionsTelemetryEvent(BaseTelemetryEvent):
    registered_functions: list[RegisteredFunction]
    name: str = 'controller_registered_functions'

@dataclass
class AgentStepTelemetryEvent(BaseTelemetryEvent):
    agent_id: str
    step: int
    step_error: list[str]
    consecutive_failures: int
    actions: list[dict]
    name: str = 'agent_step'

@dataclass
class AgentRunTelemetryEvent(BaseTelemetryEvent):
    agent_id: str
    use_vision: bool
    task: str
    model_name: str
    chat_model_library: str
    version: str
    source: str
    name: str = 'agent_run'

@dataclass
class AgentEndTelemetryEvent(BaseTelemetryEvent):
    agent_id: str
    steps: int
    max_steps_reached: bool
    #is_done: bool  # This field appears in your reconstruction but not in the bytecode
    success: bool
    #total_input_tokens: int  # This field appears in your reconstruction but not in the bytecode
    #total_duration_seconds: float  # This field appears in your reconstruction but not in the bytecode
    errors: list[str]
    name: str = 'agent_end'


import json
import os
import time
from browser_use.browser.browser import Browser, BrowserConfig

async def test_process_dom():
    browser = Browser(config=BrowserConfig(headless=False))
    async with await browser.new_context() as context:
        page = await context.get_current_page()
        await page.goto('https://kayak.com/flights')
        # await page.goto('https://google.com/flights')
        # await page.goto('https://immobilienscout24.de')
        # await page.goto('https://seleniumbase.io/w3schools/iframes')
        time.sleep(3)
        with open('browser_use/dom/buildDomTree.js', 'r') as f:
            js_code = f.read()
        start = time.time()
        dom_tree = await page.evaluate(js_code)
        end = time.time()
        # print(dom_tree)
        print(f'Time: {end - start:.2f}s')
        os.makedirs('./tmp', exist_ok=True)
        with open('./tmp/dom.json', 'w') as f:
            json.dump(dom_tree, f, indent=1)
        # both of these work for immobilienscout24.de
        # await page.click('.sc-dcJsrY.ezjNCe')
        # await page.click(
        #   'div > div:nth-of-type(2) > div > div:nth-of-type(2) > div > div:nth-of-type(2) > div > div > div > button:nth-of-type(2)'
        # )
        input('Press Enter to continue...')

if __name__ == '__main__':
    import asyncio
    asyncio.run(test_process_dom())


import asyncio
from pathlib import Path
from fastapi import FastAPI, WebSocket, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from app.logger import logger
from app.tools.system.system_vnc import SystemVncManager
from contextlib import asynccontextmanager
import subprocess
from pydantic import BaseModel
import aiohttp
import tempfile
import os
import base64

system_vnc_manager = SystemVncManager()

class XdotoolCommand(BaseModel):
    command: str
    value: str

@asynccontextmanager
async def lifespan(app: FastAPI):
    system_vnc_manager.start_all()
    yield
    system_vnc_manager.cleanup()

app = FastAPI(lifespan=lifespan)
app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

@app.websocket("/vnc")
async def vnc_websocket_endpoint(ws: WebSocket):
    """WebSocket endpoint for VNC connections"""
    await ws.accept()
    logger.info("New VNC WebSocket connection established")
    
    try:
        reader, writer = await asyncio.open_connection('localhost', 5900)
        
        async def forward_to_vnc():
            try:
                while True:
                    data = await ws.receive_bytes()
                    writer.write(data)
                    await writer.drain()
            except Exception as e:
                logger.error(f"Error forwarding to VNC: {e}")
                writer.close()
                await writer.wait_closed()
        
        async def forward_from_vnc():
            try:
                while True:
                    data = await reader.read(65536)  # 64KB buffer for VNC data
                    if not data:
                        break
                    await ws.send_bytes(data)
            except Exception as e:
                logger.error(f"Error forwarding from VNC: {e}")
                writer.close()
                await writer.wait_closed()
        
        to_vnc = asyncio.create_task(forward_to_vnc())
        from_vnc = asyncio.create_task(forward_from_vnc())
        
        done, pending = await asyncio.wait(
            [to_vnc, from_vnc],
            return_when=asyncio.FIRST_COMPLETED
        )
        
        for task in pending:
            task.cancel()
        
        await asyncio.gather(*pending, return_exceptions=True)
        
    except Exception as e:
        logger.error(f"VNC WebSocket error: {e}")
    finally:
        await ws.close()
        logger.info("VNC WebSocket connection closed")

@app.post("/control")
async def xdotool_command(command: XdotoolCommand):
    """Execute xdotool command in the virtual machine"""
    try:
        logger.info(f"Executing xdotool command: {command.command} {command.value}")
        # Execute xdotool command
        process = subprocess.Popen(
            ["xdotool", command.command, command.value],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        stdout, stderr = process.communicate()
        
        if process.returncode != 0:
            raise HTTPException(
                status_code=400,
                detail=f"xdotool command failed: {stderr.decode()}"
            )
            
        return {"status": "success", "output": stdout.decode()}
    except Exception as e:
        logger.error(f"Error executing xdotool command: {e}")
        raise HTTPException(
            status_code=500,
            detail=f"Failed to execute xdotool command: {str(e)}"
        )

@app.get("/check")
async def check():
    return {"status": "success"}

async def get_ssl() -> tuple[str | None, str | None]:
    if os.environ.get("VNC_SSL_CERT") and os.environ.get("VNC_SSL_KEY"):
        cert_file = tempfile.NamedTemporaryFile(delete=False)
        key_file = tempfile.NamedTemporaryFile(delete=False)
        cert_data = base64.b64decode(os.environ.get("VNC_SSL_CERT"))
        key_data = base64.b64decode(os.environ.get("VNC_SSL_KEY"))
        
        cert_file.write(cert_data)
        key_file.write(key_data)
        cert_file.close()
        key_file.close()
        
        return Path(cert_file.name).resolve(), Path(key_file.name).resolve()
    else:
        return Path('app/key.pem').resolve(), Path('app/cert.pem').resolve()

async def start_vnc_server(host: str = "0.0.0.0", port: int = 8081):
    """Start the VNC server"""
    import uvicorn
    
    logger.info(f"Starting VNC server on {host}:{port}")
    ssl_keyfile, ssl_certfile = await get_ssl()
    config = uvicorn.Config(
        app,
        host=host,
        port=port,
        ssl_keyfile=ssl_keyfile,
        ssl_certfile=ssl_certfile
    )
    server = uvicorn.Server(config)
    await server.serve()


"""
Playwright browser on steroids.
"""

import asyncio
import base64
import gc
import io
import json
import logging
import os
import pathlib
import re
import sys
import time
import uuid
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, TypedDict, cast
from urllib.parse import urlparse

from PIL import Image
from playwright.async_api import Browser as PlaywrightBrowser
from playwright.async_api import BrowserContext as PlaywrightBrowserContext
from playwright.async_api import Error as PlaywrightError
from playwright.async_api import Page, ElementHandle, FrameLocator
from playwright.async_api import TimeoutError as PlaywrightTimeoutError

from browser_use.browser.views import (
    BrowserError, 
    BrowserState, 
    ClickableElementData, 
    ElementInfo, 
    ExtractedPageContentInfo, 
    ScreenshotError, 
    TabInfo, 
    URLNotAllowedError
)
from browser_use.dom.views import DOMElementNode
from browser_use.dom.service import DomService
from browser_use.utils import time_execution_sync, time_execution_async

if TYPE_CHECKING:
    from browser_use.browser.browser import Browser
    from browser_use.dom.views import SelectorMap

# Constants
JS_DIR = os.path.normpath(os.path.join(os.path.dirname(__file__), '../../app_data/js'))
runExtensionActionJs = pathlib.Path(JS_DIR, 'runExtensionAction.js').read_text()
URL_BLANK = 'about:blank'

# Configure PIL logging
logging.getLogger('PIL').setLevel(logging.CRITICAL + 1)
logger = logging.getLogger(__name__)


class BrowserContextWindowSize(TypedDict):
    """
    Size configuration for browser window
    """
    width: int
    height: int


class PageDeadError(Exception):
    """
    Exception raised when a page is no longer accessible
    """
    pass


@dataclass
class BrowserContextConfig:
    """
    Configuration for the BrowserContext.

    Default values:
        cookies_file: None
            Path to cookies file for persistence

            disable_security: True
                    Disable browser security features

        minimum_wait_page_load_time: 0.5
            Minimum time to wait before getting page state for LLM input

            wait_for_network_idle_page_load_time: 1.0
                    Time to wait for network requests to finish before getting page state.
                    Lower values may result in incomplete page loads.

        maximum_wait_page_load_time: 5.0
            Maximum time to wait for page load before proceeding anyway

        wait_between_actions: 1.0
            Time to wait between multiple per step actions

        browser_window_size: {
                'width': 1280,
                'height': 1100,
            }
            Default browser window size

        no_viewport: False
            Disable viewport

        save_recording_path: None
            Path to save video recordings

        save_downloads_path: None
            Path to save downloads to

        trace_path: None
            Path to save trace files. It will auto name the file with the TRACE_PATH/{context_id}.zip

        locale: None
            Specify user locale, for example en-GB, de-DE, etc. Locale will affect navigator.language value, Accept-Language request header value as well as number and date formatting rules. If not provided, defaults to the system default locale.

        user_agent: 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36'
            custom user agent to use.

        highlight_elements: True
            Highlight elements in the DOM on the screen

        viewport_expansion: 500
            Viewport expansion in pixels. This amount will increase the number of elements which are included in the state what the LLM will see. If set to -1, all elements will be included (this leads to high token usage). If set to 0, only the elements which are visible in the viewport will be included.

        allowed_domains: None
            List of allowed domains that can be accessed. If None, all domains are allowed.
            Example: ['example.com', 'api.example.com']

        include_dynamic_attributes: bool = True
            Include dynamic attributes in the CSS selector. If you want to reuse the css_selectors, it might be better to set this to False.
    """

    cookies_file: str | None = None
    minimum_wait_page_load_time: float = 0.25
    wait_for_network_idle_page_load_time: float = 0.5
    maximum_wait_page_load_time: float = 5
    wait_between_actions: float = 0.5

    disable_security: bool = True

    browser_window_size:  BrowserContextWindowSize = field(default_factory=lambda: {
        'width': 1280 if sys.platform == 'linux' else 1600, 
        'height': 1024 if sys.platform == 'linux' else 900
    })
    no_viewport: Optional[bool] = None

    save_recording_path: str | None = None
    save_downloads_path: str | None = None
    trace_path: str | None = None
    locale: str | None = None
    user_agent: str = (
        'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/85.0.4183.102 Safari/537.36'
    )

    highlight_elements: bool = True
    viewport_expansion: int = 500
    allowed_domains: list[str] | None = None
    include_dynamic_attributes: bool = True

    _force_keep_context_alive: bool = False


@dataclass
class BrowserSession:
    """
    Browser session containing the context and cached state
    """
    context: PlaywrightBrowserContext
    cached_state: BrowserState | None


@dataclass
class BrowserContextState:
    """
    State of the browser context
    """
    target_id: str | None = None  # CDP target ID


class BrowserContext:
    """
    Enhanced browser context that wraps Playwright's browser context
    with additional functionality for browser automation.
    """
    
    def __init__(
        self,
        browser: 'Browser',
        config: BrowserContextConfig = BrowserContextConfig(),
        state: Optional[BrowserContextState] = None,
    ):
        """
        Initialize a new browser context
        
        Args:
            browser: Browser instance that this context belongs to
            config: Configuration for the browser context
            state: Optional state for the browser context
        """
        self.context_id = str(uuid.uuid4())
        logger.debug(f'Initializing new browser context with id: {self.context_id}')

        self.config = config
        self.browser = browser
        self.state = state or BrowserContextState()
        self.context = None
        
        # Initialize these as None - they'll be set up when needed
        self.session = None
        self._page_event_handler = None
        self.current_state = None

    async def __aenter__(self):
        """Async context manager entry"""
        await self._initialize_session()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit"""
        await self.close()
    
    async def ensure_page_alive(self):
        """Ensure the page is still alive"""
        if self.session is None:
            await self._initialize_session()
            return
        
        try:
            page = await self.get_current_page()
            await page.evaluate('1')  # Basic check to see if page is responsive
        except Exception:
            # Page is not alive, reinitialize
            await self._initialize_session()

    @time_execution_async('--close')
    async def close(self):
        """Close the browser instance"""
        logger.debug('Closing browser context')

        try:
            if self.session is None:
                return

            # Then remove CDP protocol listeners
            if self._page_event_handler and self.session.context:
                try:
                    # This actually sends a CDP command to unsubscribe
                    self.session.context.remove_listener('page', self._page_event_handler)
                except Exception as e:
                    logger.debug(f'Failed to remove CDP listener: {e}')
                self._page_event_handler = None

            await self.save_cookies()

            if self.config.trace_path:
                try:
                    await self.session.context.tracing.stop(path=os.path.join(self.config.trace_path, f'{self.context_id}.zip'))
                except Exception as e:
                    logger.debug(f'Failed to stop tracing: {e}')

            # This is crucial - it closes the CDP connection
            if not self.config._force_keep_context_alive:
                try:
                    await self.session.context.close()
                except Exception as e:
                    logger.debug(f'Failed to close context: {e}')

        finally:
            # Dereference everything
            self.session = None
            self._page_event_handler = None

    def __del__(self):
        """Cleanup when object is destroyed"""
        if not self.config._force_keep_context_alive and self.session is not None:
            logger.debug('BrowserContext was not properly closed before destruction')
            try:
                # Use sync Playwright method for force cleanup
                if hasattr(self.session.context, '_impl_obj'):
                    asyncio.run(self.session.context._impl_obj.close())

                self.session = None
                gc.collect()
            except Exception as e:
                logger.warning(f'Failed to force close browser context: {e}')

    @time_execution_async('--initialize_session')
    async def _initialize_session(self):
        """Initialize the browser session"""
        logger.debug('Initializing browser context')

        playwright_browser = await self.browser.get_playwright_browser()
        context = await self._create_context(playwright_browser)
        self._page_event_handler = None

        # Add listener for new page events immediately after creating context
        # self._add_new_page_listener(context)

        # Get or create a page to use
        pages = context.pages

        self.session = BrowserSession(
            context=context,
            cached_state=None,
        )

        active_page = None
        
        # if hasattr(self.browser, 'config') and getattr(self.browser.config, 'cdp_url', None):
            # If we have a saved target ID, try to find and activate it
        # if self.state.target_id:
        #     targets = await self._get_cdp_targets()
        #     for target in targets:
        #         if target['targetId'] == self.state.target_id:
        #             # Find matching page by URL
        #             for page in pages:
        #                 if page.url == target['url']:
        #                     active_page = page
        #                     break
        #             break

        # If no target ID or couldn't find it, use existing page or create new
        if not active_page:
            if pages:
                active_page = pages[0]
                logger.debug('Using existing page')
            else:
                active_page = await context.new_page()
                logger.debug('Created new page')

            # Get target ID for the active page
            # if hasattr(self.browser, 'config') and getattr(self.browser.config, 'cdp_url', None):
            # targets = await self._get_cdp_targets()
            # for target in targets:
            #     if target['url'] == active_page.url:
            #         self.state.target_id = target['targetId']
            #         break

        # Bring page to front
        await active_page.bring_to_front()
        await active_page.wait_for_load_state('load')

        return self.session

    def _add_new_page_listener(self, context: PlaywrightBrowserContext):
        """Add listener for new page events"""
        async def on_page(page: Page):
            # if hasattr(self.browser, 'config') and getattr(self.browser.config, 'cdp_url', None):
            await page.reload()  # Reload the page to avoid timeout errors
            await page.wait_for_load_state()
            logger.debug(f'New page opened: {page.url}')
            targets = await self._get_cdp_targets()
            # 筛选 type=page 的 targets
            page_targets = [target for target in targets if target.get('type') == 'page']
            if page_targets:
                self.state.target_id = page_targets[0]['targetId']
            

            self._page_event_handler = on_page
            context.on('page', on_page)

    async def get_session(self) -> BrowserSession:
        """Lazy initialization of the browser and related components"""
        if self.session is None:
            return await self._initialize_session()
        return self.session

    async def get_current_page(self) -> Page:
        """Get the current page"""
        session = await self.get_session()
        return await self._get_current_page(session)

    async def _create_context(self, browser: PlaywrightBrowser):
        """Creates a new browser context with anti-detection measures and loads cookies if available."""
        # logger.info(f'Creating context for browser: {browser.contexts}')
        context = browser.contexts[0]
              # Expose anti-detection scripts
        await context.add_init_script(
            """
            // Webdriver property
            Object.defineProperty(navigator, 'webdriver', {
                get: () => undefined
            });

            // Languages
            Object.defineProperty(navigator, 'languages', {
                get: () => ['en-US']
            });

            // Plugins
            Object.defineProperty(navigator, 'plugins', {
                get: () => [1, 2, 3, 4, 5]
            });

            // Chrome runtime
            window.chrome = { runtime: {} };
            

            // Permissions
            const originalQuery = window.navigator.permissions.query;
            window.navigator.permissions.query = (parameters) => (
                parameters.name === 'notifications' ?
                    Promise.resolve({ state: Notification.permission }) :
                    originalQuery(parameters)
            );
            (function () {
                const originalAttachShadow = Element.prototype.attachShadow;
                Element.prototype.attachShadow = function attachShadow(options) {
                    return originalAttachShadow.call(this, { ...options, mode: "open" });
                };
            })();
            """
        )
        
        if self.config.trace_path:
            await context.tracing.start(screenshots=True, snapshots=True, sources=True)

        # Load cookies if they exist
        if self.config.cookies_file and os.path.exists(self.config.cookies_file):
            with open(self.config.cookies_file, 'r') as f:
                cookies = json.load(f)
                logger.info(f'Loaded {len(cookies)} cookies from {self.config.cookies_file}')
                await context.add_cookies(cookies)

        self.context = context
        return context
        if hasattr(self.browser, 'config'):
            if getattr(self.browser.config, 'cdp_url', None) and len(browser.contexts) > 0:
                context = browser.contexts[0]
            elif getattr(self.browser.config, 'chrome_instance_path', None) and len(browser.contexts) > 0:
                # Connect to existing Chrome instance instead of creating new one
                context = browser.contexts[0]
            else:
                # Original code for creating new context
                context = await browser.new_context(
                    viewport=self.config.browser_window_size,
                    no_viewport=False,
                    user_agent=self.config.user_agent,
                    device_scale_factor=2,
                    java_script_enabled=True,
                    bypass_csp=self.config.disable_security,
                    ignore_https_errors=self.config.disable_security,
                    record_video_dir=self.config.save_recording_path,
                    record_video_size=self.config.browser_window_size,
                    locale=self.config.locale,
                )
        else:
            # Fallback if browser doesn't have the config attribute
            context = await browser.new_context(
                viewport=self.config.browser_window_size,
                no_viewport=False,
                user_agent=self.config.user_agent,
                java_script_enabled=True,
                device_scale_factor=2,
                bypass_csp=self.config.disable_security,
                ignore_https_errors=self.config.disable_security,
                record_video_dir=self.config.save_recording_path,
                record_video_size=self.config.browser_window_size,
                locale=self.config.locale,
            )

        if self.config.trace_path:
            await context.tracing.start(screenshots=True, snapshots=True, sources=True)

        # Load cookies if they exist
        if self.config.cookies_file and os.path.exists(self.config.cookies_file):
            with open(self.config.cookies_file, 'r') as f:
                cookies = json.load(f)
                logger.info(f'Loaded {len(cookies)} cookies from {self.config.cookies_file}')
                await context.add_cookies(cookies)

        # Expose anti-detection scripts
        await context.add_init_script(
            """
            // Webdriver property
            Object.defineProperty(navigator, 'webdriver', {
                get: () => undefined
            });

            // Languages
            Object.defineProperty(navigator, 'languages', {
                get: () => ['en-US']
            });

            // Plugins
            Object.defineProperty(navigator, 'plugins', {
                get: () => [1, 2, 3, 4, 5]
            });

            // Chrome runtime
            window.chrome = { runtime: {} };

            // Permissions
            const originalQuery = window.navigator.permissions.query;
            window.navigator.permissions.query = (parameters) => (
                parameters.name === 'notifications' ?
                    Promise.resolve({ state: Notification.permission }) :
                    originalQuery(parameters)
            );
            (function () {
                const originalAttachShadow = Element.prototype.attachShadow;
                Element.prototype.attachShadow = function attachShadow(options) {
                    return originalAttachShadow.call(this, { ...options, mode: "open" });
                };
            })();
            """
        )

        return context

    async def _wait_for_stable_network(self):
        """Wait for network to become stable"""
        page = await self.get_current_page()

        pending_requests = set()
        last_activity = asyncio.get_event_loop().time()

        # Define relevant resource types and content types
        RELEVANT_RESOURCE_TYPES = {
            'document',
            'stylesheet',
            'image',
            'font',
            'script',
            'iframe',
        }

        RELEVANT_CONTENT_TYPES = {
            'text/html',
            'text/css',
            'application/javascript',
            'image/',
            'font/',
            'application/json',
        }

        # Additional patterns to filter out
        IGNORED_URL_PATTERNS = {
            # Analytics and tracking
            'analytics',
            'tracking',
            'telemetry',
            'beacon',
            'metrics',
            # Ad-related
            'doubleclick',
            'adsystem',
            'adserver',
            'advertising',
            # Social media widgets
            'facebook.com/plugins',
            'platform.twitter',
            'linkedin.com/embed',
            # Live chat and support
            'livechat',
            'zendesk',
            'intercom',
            'crisp.chat',
            'hotjar',
            # Push notifications
            'push-notifications',
            'onesignal',
            'pushwoosh',
            # Background sync/heartbeat
            'heartbeat',
            'ping',
            'alive',
            # WebRTC and streaming
            'webrtc',
            'rtmp://',
            'wss://',
            # Common CDNs for dynamic content
            'cloudfront.net',
            'fastly.net',
        }

        async def on_request(request):
            # Filter by resource type
            if request.resource_type not in RELEVANT_RESOURCE_TYPES:
                return

            # Filter out streaming, websocket, and other real-time requests
            if request.resource_type in {
                'websocket',
                'media',
                'eventsource',
                'manifest',
                'other',
            }:
                return

            # Filter out by URL patterns
            url = request.url.lower()
            if any(pattern in url for pattern in IGNORED_URL_PATTERNS):
                return

            # Filter out data URLs and blob URLs
            if url.startswith(('data:', 'blob:')):
                return

            # Filter out requests with certain headers
            headers = request.headers
            if headers.get('purpose') == 'prefetch' or headers.get('sec-fetch-dest') in [
                'video',
                'audio',
            ]:
                return

            nonlocal last_activity
            pending_requests.add(request)
            last_activity = asyncio.get_event_loop().time()
            # logger.debug(f'Request started: {request.url} ({request.resource_type})')

        async def on_response(response):
            request = response.request
            if request not in pending_requests:
                return

            # Filter by content type if available
            content_type = response.headers.get('content-type', '').lower()

            # Skip if content type indicates streaming or real-time data
            if any(
                t in content_type
                for t in [
                    'streaming',
                    'video',
                    'audio',
                    'webm',
                    'mp4',
                    'event-stream',
                    'websocket',
                    'protobuf',
                ]
            ):
                pending_requests.remove(request)
                return

            # Only process relevant content types
            if not any(ct in content_type for ct in RELEVANT_CONTENT_TYPES):
                pending_requests.remove(request)
                return

            # Skip if response is too large (likely not essential for page load)
            content_length = response.headers.get('content-length')
            if content_length and int(content_length) > 5 * 1024 * 1024:  # 5MB
                pending_requests.remove(request)
                return

            nonlocal last_activity
            pending_requests.remove(request)
            last_activity = asyncio.get_event_loop().time()
            # logger.debug(f'Request resolved: {request.url} ({content_type})')

        # Attach event listeners
        page.on('request', on_request)
        page.on('response', on_response)

        try:
            # Wait for idle time
            start_time = asyncio.get_event_loop().time()
            while True:
                await asyncio.sleep(0.1)
                now = asyncio.get_event_loop().time()
                if len(pending_requests) == 0 and (now - last_activity) >= self.config.wait_for_network_idle_page_load_time:
                    break
                if now - start_time > self.config.maximum_wait_page_load_time:
                    logger.debug(
                        f'Network timeout after {self.config.maximum_wait_page_load_time}s with {len(pending_requests)} '
                        f'pending requests: {[r.url for r in pending_requests]}'
                    )
                    break

        finally:
            # Clean up event listeners
            page.remove_listener('request', on_request)
            page.remove_listener('response', on_response)

        logger.debug(f'Network stabilized for {self.config.wait_for_network_idle_page_load_time} seconds')

    async def _wait_for_page_and_frames_load(self, timeout_overwrite: float | None = None):
        """
        Ensures page is fully loaded before continuing.
        Waits for either network to be idle or minimum WAIT_TIME, whichever is longer.
        Also checks if the loaded URL is allowed.
        """
        # Start timing
        start_time = time.time()

        # Wait for page load
        try:
            await self._wait_for_stable_network()

            # Check if the loaded URL is allowed
            page = await self.get_current_page()
            await self._check_and_handle_navigation(page)
        except URLNotAllowedError as e:
            raise e
        except Exception:
            logger.warning('Page load failed, continuing...')
            pass

        # Calculate remaining time to meet minimum WAIT_TIME
        elapsed = time.time() - start_time
        remaining = max((timeout_overwrite or self.config.minimum_wait_page_load_time) - elapsed, 0)

        logger.debug(f'--Page loaded in {elapsed:.2f} seconds, waiting for additional {remaining:.2f} seconds')

        # Sleep remaining time if needed
        if remaining > 0:
            await asyncio.sleep(remaining)

    def _is_url_allowed(self, url: str) -> bool:
        """Check if a URL is allowed based on the whitelist configuration."""
        if not self.config.allowed_domains:
            return True

        try:
            parsed_url = urlparse(url)
            domain = parsed_url.netloc.lower()

            # Remove port number if present
            if ':' in domain:
                domain = domain.split(':')[0]

            # Check if domain matches any allowed domain pattern
            return any(
                domain == allowed_domain.lower() or domain.endswith('.' + allowed_domain.lower())
                for allowed_domain in self.config.allowed_domains
            )
        except Exception as e:
            logger.error(f'Error checking URL allowlist: {str(e)}')
            return False

    async def _check_and_handle_navigation(self, page: Page) -> None:
        """Check if current page URL is allowed and handle if not."""
        if not self._is_url_allowed(page.url):
            logger.warning(f'Navigation to non-allowed URL detected: {page.url}')
            try:
                await self.go_back()
            except Exception as e:
                logger.error(f'Failed to go back after detecting non-allowed URL: {str(e)}')
            raise URLNotAllowedError(f'Navigation to non-allowed URL: {page.url}')

    async def navigate_to(self, url: str):
        """Navigate to a URL"""
        if not self._is_url_allowed(url):
            raise BrowserError(f'Navigation to non-allowed URL: {url}')

        page = await self.get_current_page()
        await page.goto(url)
        await page.wait_for_load_state()

    async def refresh_page(self):
        """Refresh the current page"""
        page = await self.get_current_page()
        await page.reload()
        await page.wait_for_load_state()

    async def go_back(self):
        """Navigate back in history"""
        page = await self.get_current_page()
        try:
            # 10 ms timeout
            await page.go_back(timeout=10, wait_until='domcontentloaded')
            # await self._wait_for_page_and_frames_load(timeout_overwrite=1.0)
        except Exception as e:
            # Continue even if its not fully loaded, because we wait later for the page to load
            logger.debug(f'During go_back: {e}')

    async def go_forward(self):
        """Navigate forward in history"""
        page = await self.get_current_page()
        try:
            await page.go_forward(timeout=10, wait_until='domcontentloaded')
        except Exception as e:
            # Continue even if its not fully loaded, because we wait later for the page to load
            logger.debug(f'During go_forward: {e}')

    async def close_current_tab(self):
        """Close the current tab"""
        session = await self.get_session()
        page = await self._get_current_page(session)
        await page.close()

        # Switch to the first available tab if any exist
        if session.context.pages:
            await self.switch_to_tab(0)

        # otherwise the browser will be closed

    async def get_page_html(self) -> str:
        """Get the current page HTML content"""
        page = await self.get_current_page()
        return await page.content()

    async def execute_javascript(self, script: str):
        """Execute JavaScript code on the page"""
        page = await self.get_current_page()
        return await page.evaluate(script)

    @time_execution_sync('--get_state')
    async def get_state(self) -> BrowserState:
        """Get the current state of the browser"""
        await self._wait_for_page_and_frames_load()
        session = await self.get_session()
        session.cached_state = await self._update_state()

        # Save cookies if a file is specified
        if self.config.cookies_file:
            asyncio.create_task(self.save_cookies())

        return session.cached_state

    async def _update_state(self, focus_element: int = -1) -> BrowserState:
        """Update and return state."""
        session = await self.get_session()

        # Check if current page is still valid, if not switch to another available page
        try:
            page = await self.get_current_page()
            # Test if page is still accessible
            await page.evaluate('1')
        except Exception as e:
            logger.debug(f'Current page is no longer accessible: {str(e)}')
            # Get all available pages
            pages = session.context.pages
            if pages:
                self.state.target_id = None
                page = await self._get_current_page(session)
                logger.debug(f'Switched to page: {await page.title()}')
            else:
                raise BrowserError('Browser closed: no valid pages available')

        try:
            await self.remove_highlights()
            dom_service = DomService(page)
            content = await dom_service.get_clickable_elements(
                focus_element=focus_element,
                viewport_expansion=self.config.viewport_expansion,
                highlight_elements=self.config.highlight_elements,
            )
            
            screenshot_b64 = await self.take_screenshot()
            pixels_above, pixels_below = await self.get_scroll_info(page)

            self.current_state = BrowserState(
                element_tree=content.element_tree,
                selector_map=content.selector_map,
                url=page.url,
                title=await page.title(),
                tabs=await self.get_tabs_info(),
                screenshot=screenshot_b64,
                pixels_above=pixels_above,
                pixels_below=pixels_below,
            )

            return self.current_state
        except Exception as e:
            logger.error(f'Failed to update state: {str(e)}')
            # Return last known good state if available
            if hasattr(self, 'current_state'):
                return self.current_state
            raise

    async def extract_content_info(self):
        """Extract markdown from the page"""
        page = await self.get_current_page()
        try:
            # Extract the main content
            result = await page.evaluate("""
            function extractContent() {
                // Extract main article content
                const article = document.querySelector('article');
                if (article) {
                    return article.innerText;
                }
                
                // Try common content containers
                const main = document.querySelector('main');
                if (main) {
                    return main.innerText;
                }
                
                // Fallback to body content
                return document.body.innerText;
            }
            extractContent();
            """)
            
            return ExtractedPageContentInfo(
                content=result,
                url=page.url,
                title=await page.title()
            )
        except Exception as e:
            logger.error(f"Failed to extract content: {e}")
            return ExtractedPageContentInfo(
                content="",
                url=page.url,
                title=await page.title()
            )

    async def progressive_wait_for_load(self, timeout_ms: int = 30000):
        """
        分阶段等待页面加载
        先等待 DOM 内容加载，再等待完全加载
        如果任一阶段超时，继续执行但返回对应状态
        @returns is_dom_ready, is_pdf
        """
        page = await self.get_current_page()
        is_dom_ready = False
        is_pdf = False
        
        try:
            # Check if this is a PDF
            is_pdf = await page.evaluate("""
                document.querySelector('embed[type="application/pdf"]') != null ||
                document.querySelector('object[type="application/pdf"]') != null ||
                window.location.href.toLowerCase().endsWith('.pdf')
            """)
            
            if is_pdf:
                logger.debug("PDF detected, using minimal wait")
                await asyncio.sleep(1)  # Minimal wait for PDF
                return True, True
                
            # First wait for DOMContentLoaded
            try:
                await page.wait_for_load_state("domcontentloaded", timeout=timeout_ms/2)
                is_dom_ready = True
            except PlaywrightTimeoutError:
                logger.warning("DOMContentLoaded timeout")
                return False, False
                
            # Then wait for full load
            try:
                await page.wait_for_load_state("load", timeout=timeout_ms/2)
            except PlaywrightTimeoutError:
                logger.warning("Full page load timeout")
                # Continue with DOM content at least
            
            return is_dom_ready, is_pdf
            
        except Exception as e:
            logger.error(f"Error during progressive load wait: {e}")
            return is_dom_ready, is_pdf

    async def _wait_page_content_load(self):
        """Wait for page content to load"""
        page = await self.get_current_page()
        try:
            # Wait for any potential AJAX or dynamic content
            await asyncio.sleep(1)
            
            # Check if page is waiting on user input
            is_waiting = await page.evaluate("""
                document.querySelector('input:required:invalid') != null ||
                document.querySelector('form:invalid') != null
            """)
            
            if is_waiting:
                logger.debug("Page requires user input, stopping wait")
                return
                
            # Try to detect infinite loaders
            start_time = time.time()
            previous_html = await page.content()
            
            # Check for content stabilization
            for _ in range(3):
                await asyncio.sleep(0.5)
                current_html = await page.content()
                
                if current_html == previous_html:
                    logger.debug("Content stabilized")
                    break
                    
                previous_html = current_html
                
                # If we've waited too long, stop
                if time.time() - start_time > 5:
                    logger.debug("Content wait timeout")
                    break
                    
        except Exception as e:
            logger.error(f"Error waiting for page content: {e}")

    @time_execution_async('--take_screenshot')
    async def take_screenshot(self, marked: bool = False, save_path: str = None) -> str:
        """
        Returns a base64 encoded screenshot of the current page.
        """
        page = await self.get_current_page()

        await page.bring_to_front()
        await page.wait_for_load_state()

        screenshot = await page.screenshot(
            full_page=False,
            animations='disabled',
            # quality=80,
            # scale='css',
            # type='jpeg'
        )

        logger.debug(f'----------------{save_path}')
        if save_path:
            try:
                os.makedirs(os.path.dirname(save_path), exist_ok=True)
                with open(save_path, 'wb') as f:
                    f.write(screenshot)
                logger.debug(f'截图已保存到 {save_path}')
            except Exception as e:
                logger.error(f'保存截图失败: {e}')
                
        screenshot_b64 = base64.b64encode(screenshot).decode('utf-8')

        return screenshot_b64

    @time_execution_async('--remove_highlights')
    async def remove_highlights(self):
        """
        Removes all highlight overlays and labels created by the highlightElement function.
        Handles cases where the page might be closed or inaccessible.
        """
        try:
            page = await self.get_current_page()
            await page.evaluate(
                """
                try {
                    // Remove the highlight container and all its contents
                    const container = document.getElementById('playwright-highlight-container');
                    if (container) {
                        container.remove();
                    }

                    // Remove highlight attributes from elements
                    const highlightedElements = document.querySelectorAll('[browser-user-highlight-id^="playwright-highlight-"]');
                    highlightedElements.forEach(el => {
                        el.removeAttribute('browser-user-highlight-id');
                    });
                } catch (e) {
                    console.error('Failed to remove highlights:', e);
                }
                """
            )
        except Exception as e:
            logger.debug(f'Failed to remove highlights (this is usually ok): {str(e)}')
            # Don't raise the error since this is not critical functionality
            pass

    @classmethod
    def _convert_simple_xpath_to_css_selector(cls, xpath: str) -> str:
        """Converts simple XPath expressions to CSS selectors."""
        if not xpath:
            return ''

        # Remove leading slash if present
        xpath = xpath.lstrip('/')

        # Split into parts
        parts = xpath.split('/')
        css_parts = []

        for part in parts:
            if not part:
                continue

            # Handle index notation [n]
            if '[' in part:
                base_part = part[: part.find('[')]
                index_part = part[part.find('[') :]

                # Handle multiple indices
                indices = [i.strip('[]') for i in index_part.split(']')[:-1]]

                for idx in indices:
                    try:
                        # Handle numeric indices
                        if idx.isdigit():
                            index = int(idx) - 1
                            base_part += f':nth-of-type({index + 1})'
                        # Handle last() function
                        elif idx == 'last()':
                            base_part += ':last-of-type'
                        # Handle position() functions
                        elif 'position()' in idx:
                            if '>1' in idx:
                                base_part += ':nth-of-type(n+2)'
                    except ValueError:
                        continue

                css_parts.append(base_part)
            else:
                css_parts.append(part)

        base_selector = ' > '.join(css_parts)
        return base_selector

    @classmethod
    @time_execution_sync('--enhanced_css_selector_for_element')
    def _enhanced_css_selector_for_element(cls, element: DOMElementNode, include_dynamic_attributes: bool = True) -> str:
        """
        Creates a CSS selector for a DOM element, handling various edge cases and special characters.

        Args:
            element: The DOM element to create a selector for
            include_dynamic_attributes: Whether to include dynamic attributes in the selector

        Returns:
            A valid CSS selector string
        """
        try:
            # Get base selector from XPath
            css_selector = cls._convert_simple_xpath_to_css_selector(element.xpath)

            # Handle class attributes
            if 'class' in element.attributes and element.attributes['class'] and include_dynamic_attributes:
                # Define a regex pattern for valid class names in CSS
                valid_class_name_pattern = re.compile(r'^[a-zA-Z_][a-zA-Z0-9_-]*')

                # Iterate through the class attribute values
                classes = element.attributes['class'].split()
                for class_name in classes:
                    # Skip empty class names
                    if not class_name.strip():
                        continue

                    # Check if the class name is valid
                    if valid_class_name_pattern.match(class_name):
                        # Append the valid class name to the CSS selector
                        css_selector += f'.{class_name}'
                    else:
                        # Skip invalid class names
                        continue

            # Expanded set of safe attributes that are stable and useful for selection
            SAFE_ATTRIBUTES = {
                # Data attributes (if they're stable in your application)
                'id',
                # Standard HTML attributes
                'name',
                'type',
                'placeholder',
                # Accessibility attributes
                'aria-label',
                'aria-labelledby',
                'aria-describedby',
                'role',
                # Common form attributes
                'for',
                'autocomplete',
                'required',
                'readonly',
                # Media attributes
                'alt',
                'title',
                'src',
                # Custom stable attributes (add any application-specific ones)
                'href',
                'target',
            }

            if include_dynamic_attributes:
                dynamic_attributes = {
                    'data-id',
                    'data-qa',
                    'data-cy',
                    'data-testid',
                }
                SAFE_ATTRIBUTES.update(dynamic_attributes)

            # Handle other attributes
            for attribute, value in element.attributes.items():
                if attribute == 'class':
                    continue

                # Skip invalid attribute names
                if not attribute.strip():
                    continue

                if attribute not in SAFE_ATTRIBUTES:
                    continue

                # Escape special characters in attribute names
                safe_attribute = attribute.replace(':', r'\:')

                # Handle different value cases
                if value == '':
                    css_selector += f'[{safe_attribute}]'
                elif any(char in value for char in '"\'<>`\n\r\t'):
                    # Use contains for values with special characters
                    # Regex-substitute *any* whitespace with a single space, then strip.
                    collapsed_value = re.sub(r'\s+', ' ', value).strip()
                    # Escape embedded double-quotes.
                    safe_value = collapsed_value.replace('"', '\\"')
                    css_selector += f'[{safe_attribute}*="{safe_value}"]'
                else:
                    css_selector += f'[{safe_attribute}="{value}"]'

            return css_selector

        except Exception:
            # Fallback to a more basic selector if something goes wrong
            tag_name = element.tag_name or '*'
            return f"{tag_name}[highlight_index='{element.highlight_index}']"

    @time_execution_async('--get_locate_element')
    async def get_locate_element(self, element: DOMElementNode) -> Optional[ElementHandle]:
        """
        Locate an element in the DOM using the enhanced CSS selector
        
        Args:
            element: The DOM element to locate
            
        Returns:
            The element handle if found, None otherwise
        """
        current_frame = await self.get_current_page()

        # Start with the target element and collect all parents
        parents: list[DOMElementNode] = []
        current = element
        while current.parent is not None:
            parent = current.parent
            parents.append(parent)
            current = parent

        # Reverse the parents list to process from top to bottom
        parents.reverse()

        # Process all iframe parents in sequence
        iframes = [item for item in parents if item.tag_name == 'iframe']
        for parent in iframes:
            css_selector = self._enhanced_css_selector_for_element(
                parent,
                include_dynamic_attributes=self.config.include_dynamic_attributes,
            )
            current_frame = current_frame.frame_locator(css_selector)

        css_selector = self._enhanced_css_selector_for_element(
            element, include_dynamic_attributes=self.config.include_dynamic_attributes
        )

        try:
            if isinstance(current_frame, FrameLocator):
                element_handle = await current_frame.locator(css_selector).element_handle()
                return element_handle
            else:
                # Try to scroll into view if hidden
                element_handle = await current_frame.query_selector(css_selector)
                if element_handle:
                    await element_handle.scroll_into_view_if_needed()
                    return element_handle
                return None
        except Exception as e:
            logger.error(f'Failed to locate element: {str(e)}')
            return None

    async def get_element(self, index: int) -> ElementHandle:
        """
        Get an element handle by its index
        
        Args:
            index: The index of the element to get
            
        Returns:
            The element handle
        """
        selector_map = await self.get_selector_map()
        if index not in selector_map:
            raise BrowserError(f'No element found with index {index}')
            
        element_node = selector_map[index]
        element_handle = await self.get_locate_element(element_node)
        
        if element_handle is None:
            raise BrowserError(f'Element with index {index} not found in DOM')
            
        return element_handle

    @time_execution_async('--input_text_element_node')
    async def _input_text_element_node(self, element_node: DOMElementNode, text: str):
        """
        Input text into an element with proper error handling and state management.
        Handles different types of input fields and ensures proper element state before input.
        
        Args:
            element_node: The DOM element to input text into
            text: The text to input
        """
        try:
            element_handle = await self.get_locate_element(element_node)

            if element_handle is None:
                raise BrowserError(f'Element: {repr(element_node)} not found')

            # Ensure element is ready for input
            try:
                await element_handle.wait_for_element_state('stable', timeout=1000)
                await element_handle.scroll_into_view_if_needed(timeout=1000)
            except Exception:
                pass

            # Get element properties to determine input method
            is_contenteditable = await element_handle.get_property('isContentEditable')

            # Different handling for contenteditable vs input fields
            if await is_contenteditable.json_value():
                await element_handle.evaluate('el => el.textContent = ""')
                await element_handle.type(text, delay=5)
            else:
                await element_handle.fill(text)

        except Exception as e:
            logger.debug(f'Failed to input text into element: {repr(element_node)}. Error: {str(e)}')
            raise BrowserError(f'Failed to input text into index {element_node.highlight_index}')

    async def input_text_to_element(self, index: int, text: str, delay: float = 0):
        """
        Input text into an element using its index
        
        Args:
            index: The index of the element to input text into
            text: The text to input
            delay: Optional delay before inputting text (in seconds)
        """
        selector_map = await self.get_selector_map()
        if index not in selector_map:
            raise BrowserError(f'No element found with index {index}')
            
        element_node = selector_map[index]
        
        if delay > 0:
            await asyncio.sleep(delay)
            
        await self._input_text_element_node(element_node, text)

    @time_execution_async('--click_element_node')
    async def _click_element_node(self, element_node: DOMElementNode) -> Optional[str]:
        """
        Optimized method to click an element
        
        Args:
            element_node: The DOM element to click
            
        Returns:
            Optional download path if a download was triggered
        """
        page = await self.get_current_page()

        try:
            element_handle = await self.get_locate_element(element_node)

            if element_handle is None:
                raise Exception(f'Element: {repr(element_node)} not found')

            async def perform_click(click_func):
                """Performs the actual click, handling both download
                and navigation scenarios."""
                if self.config.save_downloads_path:
                    try:
                        # Try short-timeout expect_download to detect a file download has been been triggered
                        async with page.expect_download(timeout=5000) as download_info:
                            await click_func()
                        download = await download_info.value
                        # Determine file path
                        suggested_filename = download.suggested_filename
                        unique_filename = await self._get_unique_filename(self.config.save_downloads_path, suggested_filename)
                        download_path = os.path.join(self.config.save_downloads_path, unique_filename)
                        await download.save_as(download_path)
                        logger.debug(f'Download triggered. Saved file to: {download_path}')
                        return download_path
                    except PlaywrightTimeoutError:
                        # If no download is triggered, treat as normal click
                        logger.debug('No download triggered within timeout. Checking navigation...')
                        await page.wait_for_load_state()
                        await self._check_and_handle_navigation(page)
                else:
                    # Standard click logic if no download is expected
                    await click_func()
                    await page.wait_for_load_state()
                    await self._check_and_handle_navigation(page)

            try:
                return await perform_click(lambda: element_handle.click(timeout=1500))
            except URLNotAllowedError as e:
                raise e
            except Exception:
                try:
                    return await perform_click(lambda: page.evaluate('(el) => el.click()', element_handle))
                except URLNotAllowedError as e:
                    raise e
                except Exception as e:
                    raise Exception(f'Failed to click element: {str(e)}')

        except URLNotAllowedError as e:
            raise e
        except Exception as e:
            raise Exception(f'Failed to click element: {repr(element_node)}. Error: {str(e)}')

    async def click_element(self, index: int):
        """
        Click an element using its index
        
        Args:
            index: The index of the element to click
        """
        selector_map = await self.get_selector_map()
        if index not in selector_map:
            raise BrowserError(f'No element found with index {index}')
            
        element_node = selector_map[index]
        return await self._click_element_node(element_node)

    @time_execution_async('--get_tabs_info')
    async def get_tabs_info(self) -> list[TabInfo]:
        """
        Get information about all tabs
        
        Returns:
            List of tab info objects
        """
        session = await self.get_session()

        tabs_info = []
        for page_id, page in enumerate(session.context.pages):
            tab_info = TabInfo(page_id=page_id, url=page.url, title=await page.title())
            tabs_info.append(tab_info)

        return tabs_info

    @time_execution_async('--switch_to_tab')
    async def switch_to_tab(self, page_id: int) -> None:
        """
        Switch to a specific tab by its page_id

        @You can also use negative indices to switch to tabs from the end (Pure pythonic way)
        """
        session = await self.get_session()
        pages = session.context.pages

        # Handle negative indices
        if page_id < 0:
            page_id = len(pages) + page_id

        if page_id >= len(pages) or page_id < 0:
            raise BrowserError(f'No tab found with page_id: {page_id}')

        page = pages[page_id]

        # Check if the tab's URL is allowed before switching
        if not self._is_url_allowed(page.url):
            raise BrowserError(f'Cannot switch to tab with non-allowed URL: {page.url}')

        # Update target ID if using CDP
        # if hasattr(self.browser, 'config') and getattr(self.browser.config, 'cdp_url', None):
        # targets = await self._get_cdp_targets()
        # for target in targets:
        #     if target['url'] == page.url:
        #         self.state.target_id = target['targetId']
        #         break

        await page.bring_to_front()
        await page.wait_for_load_state()

    @time_execution_async('--create_new_tab')
    async def create_new_tab(self, url: str | None = None) -> None:
        """
        Create a new tab and optionally navigate to a URL
        
        Args:
            url: Optional URL to navigate to
        """
        if url and not self._is_url_allowed(url):
            raise BrowserError(f'Cannot create new tab with non-allowed URL: {url}')

        session = await self.get_session()
        new_page = await session.context.new_page()
        await new_page.wait_for_load_state()

        if url:
            await new_page.goto(url)
            await self._wait_for_page_and_frames_load(timeout_overwrite=1)

        # # Get target ID for new page if using CDP
        # # if hasattr(self.browser, 'config') and getattr(self.browser.config, 'cdp_url', None):
        # targets = await self._get_cdp_targets()
        # # 筛选 type=page 的 targets
        # page_targets = [target for target in targets if target.get('type') == 'page']
        # logger.info(f'Targets-------------: {targets}')
        # logger.info(f'Page Targets-------------: {page_targets}')
        
        # if page_targets:
        #     self.state.target_id = page_targets[0]['targetId']
        # elif targets:
        #     self.state.target_id = targets[0]['targetId']
        # for target in targets:
        #     if target['url'] == new_page.url:
                
        #         break

    # Helper methods
    async def _get_current_page(self, session: BrowserSession) -> Page:
        """
        Get the current page
        
        Args:
            session: The browser session
            
        Returns:
            The current page
        """
        pages = session.context.pages
        
        # cdp_session = await pages[0].context.new_cdp_session(pages[0])
        # result = await cdp_session.send('Target.getTargets')
        # await cdp_session.detach()
        # targets = result.get('targetInfos', [])
        # active_page = None
        # for target in targets:
        #     if target['type'] == 'page' and target['attached']:
        #         logger.info(f'target----: {target}')
        #         if target.get('isPage', False) and target.get('isInitialPage', False):
        #                 for page in session.context.pages:
        #                     logger.info(f'targetPage: {page.url} {target["url"]}')
        #                     if page.url == target['url']:
        #                         active_page = page
        #                         break
        # logger.info(f'Active page: {active_page}')
        # Try to find page by target ID if using CDP
        # if hasattr(self.browser, 'config') and getattr(self.browser.config, 'cdp_url', None) and self.state.target_id:
        # targets = await self._get_cdp_targets()
        # for target in targets:
        #     if target['targetId'] == self.state.target_id:
        #         for page in pages:
        #             if page.url == target['url']:
        #                 return page

        # Fallback to last page
        return pages[-1] if pages else await session.context.new_page()

    async def get_selector_map(self) -> 'SelectorMap':
        """
        Get the selector map from the cached state
        
        Returns:
            The selector map
        """
        session = await self.get_session()
        if session.cached_state is None:
            return {}
        return session.cached_state.selector_map

    async def get_element_by_index(self, index: int) -> ElementHandle | None:
        """
        Get an element handle by its index
        
        Args:
            index: The index of the element to get
            
        Returns:
            The element handle, or None if not found
        """
        selector_map = await self.get_selector_map()
        if index not in selector_map:
            return None
            
        element_handle = await self.get_locate_element(selector_map[index])
        return element_handle

    async def get_dom_element_by_index(self, index: int) -> DOMElementNode:
        """
        Get a DOM element by its index
        
        Args:
            index: The index of the element to get
            
        Returns:
            The DOM element
        """
        selector_map = await self.get_selector_map()
        if index not in selector_map:
            raise BrowserError(f'No element found with index {index}')
            
        return selector_map[index]

    async def save_cookies(self):
        """Save current cookies to file"""
        if self.session and self.session.context and self.config.cookies_file:
            try:
                cookies = await self.session.context.cookies()
                logger.debug(f'Saving {len(cookies)} cookies to {self.config.cookies_file}')

                # Check if the path is a directory and create it if necessary
                dirname = os.path.dirname(self.config.cookies_file)
                if dirname:
                    os.makedirs(dirname, exist_ok=True)

                with open(self.config.cookies_file, 'w') as f:
                    json.dump(cookies, f)
            except Exception as e:
                logger.warning(f'Failed to save cookies: {str(e)}')

    async def is_file_uploader(self, element_node: DOMElementNode, max_depth: int = 3, current_depth: int = 0) -> bool:
        """
        Check if element or its children are file uploaders
        
        Args:
            element_node: The element to check
            max_depth: Maximum depth to check children
            current_depth: Current depth of recursion
            
        Returns:
            True if the element is a file uploader, False otherwise
        """
        if current_depth > max_depth:
            return False

        # Check current element
        is_uploader = False

        if not isinstance(element_node, DOMElementNode):
            return False

        # Check for file input attributes
        if element_node.tag_name == 'input':
            is_uploader = element_node.attributes.get('type') == 'file' or element_node.attributes.get('accept') is not None

        if is_uploader:
            return True

        # Recursively check children
        if element_node.children and current_depth < max_depth:
            for child in element_node.children:
                if isinstance(child, DOMElementNode):
                    if await self.is_file_uploader(child, max_depth, current_depth + 1):
                        return True

        return False

    async def get_scroll_info(self, page: Page) -> tuple[int, int]:
        """
        Get scroll position information for the current page
        
        Args:
            page: The page to get scroll info for
            
        Returns:
            Tuple of (pixels_above, pixels_below)
        """
        scroll_y = await page.evaluate('window.scrollY')
        viewport_height = await page.evaluate('window.innerHeight')
        total_height = await page.evaluate('document.documentElement.scrollHeight')
        pixels_above = scroll_y
        pixels_below = total_height - (scroll_y + viewport_height)
        return pixels_above, pixels_below

    async def reset_context(self):
        """
        Reset the browser session
        Call this when you don't want to kill the context but just kill the state
        """
        # close all tabs and clear cached state
        session = await self.get_session()

        pages = session.context.pages
        for page in pages:
            await page.close()

        session.cached_state = None
        self.state.target_id = None

    async def show_click_visual_effect(self, x: int, y: int):
        """Show visual effect for clicking at coordinates"""
        page = await self.get_current_page()
        await page.evaluate("""
        (x, y) => {
            const clickEffect = document.createElement('div');
            clickEffect.style.position = 'absolute';
            clickEffect.style.width = '20px';
            clickEffect.style.height = '20px';
            clickEffect.style.borderRadius = '50%';
            clickEffect.style.backgroundColor = 'rgba(255, 0, 0, 0.5)';
            clickEffect.style.left = (x - 10) + 'px';
            clickEffect.style.top = (y - 10) + 'px';
            clickEffect.style.zIndex = '10000';
            clickEffect.style.pointerEvents = 'none';
            document.body.appendChild(clickEffect);
            
            // Animation
            setTimeout(() => {
                clickEffect.style.transition = 'all 0.5s ease-out';
                clickEffect.style.transform = 'scale(2)';
                clickEffect.style.opacity = '0';
                setTimeout(() => {
                    clickEffect.remove();
                }, 500);
            }, 50);
        }
        """, x, y)

    def _get_initial_state(self, page: Page) -> BrowserState:
        """
        Get the initial state of the browser
        
        Args:
            page: The current page
            
        Returns:
            The initial browser state
        """
        return BrowserState(
            element_tree=self._createRootNode(), 
            selector_map={}, 
            clickable_elements={}, 
            article_markdown='', 
            url=page.url if page else '', 
            title='', 
            tabs=[]
        )

    def _createRootNode(self) -> DOMElementNode:
        """
        Create the root DOM node
        
        Returns:
            The root DOM node
        """
        return DOMElementNode(
            False, 
            None, 
            '', 
            '', 
            {}, 
            []
        )

    async def _get_unique_filename(self, directory, filename):
        """
        Generate a unique filename by appending (1), (2), etc., if a file already exists
        
        Args:
            directory: The directory to check
            filename: The original filename
            
        Returns:
            A unique filename
        """
        base, ext = os.path.splitext(filename)
        counter = 1
        new_filename = filename
        while os.path.exists(os.path.join(directory, new_filename)):
            new_filename = f'{base} ({counter}){ext}'
            counter += 1
        return new_filename

    async def _get_cdp_targets(self) -> list[dict]:
        """
        Get all CDP targets directly using CDP protocol
        
        Returns:
            List of CDP targets
        """
        # if not hasattr(self.browser, 'config') or not getattr(self.browser.config, 'cdp_url', None) or not self.session:
        #     return []

        try:
            pages = self.session.context.pages
            if not pages:
                return []

            cdp_session = await pages[0].context.new_cdp_session(pages[0])
            result = await cdp_session.send('Target.getTargets')
            await cdp_session.detach()
            return result.get('targetInfos', [])
        except Exception as e:
            logger.debug(f'Failed to get CDP targets: {e}')
            return []

    async def _run_extension_action(self, action: str, data: dict = None, timeout: int = 30000):
        """
        Evaluates a function in the browser extension.
        """
        page = await self.get_current_page()
        try:
            result = await page.evaluate(
                runExtensionActionJs,
                {"action": action, "data": data or {}, "timeout": timeout}
            )
            return result
        except Exception as e:
            logger.error(f"Extension action '{action}' failed: {e}")
            raise BrowserError(f"Extension action failed: {str(e)}")


from typing import Optional
from pydantic import BaseModel, model_validator

# Action Input Models
class SearchBaiduAction(BaseModel):
    query: str

class GoToUrlAction(BaseModel):
    url: str

class ClickElementAction(BaseModel):
    index: int
    xpath: Optional[str] = None

class ClickByPositionAction(BaseModel):
    x: int
    y: int

class InputTextAction(BaseModel):
    index: int
    text: str
    xpath: Optional[str] = None

class ExtractPageContentAction(BaseModel):
    # This class doesn't have fields in the bytecode
    # It appears to be a model with no required fields
    pass

class DoneAction(BaseModel):
    text: str

class SwitchTabAction(BaseModel):
    page_id: int

class OpenTabAction(BaseModel):
    url: str
    
class GetAllTabsAction(BaseModel):
    pass

class ScrollAction(BaseModel):
    amount: Optional[int] = None
    index: Optional[int] = None

class SendKeysAction(BaseModel):
    keys: str

class NoParamsAction(BaseModel):
    """
    Accepts absolutely anything in the incoming data
    and discards it, so the final parsed model is empty.
    """
    @model_validator(mode='before')
    def ignore_all_inputs(cls, values):
        # No matter what the user sends, discard it and return empty.
        return {}
    
    class Config:
        extra = 'allow'


"""
Playwright browser on steroids.
"""
import asyncio
import logging
from dataclasses import dataclass, field
from playwright._impl._api_structures import ProxySettings
from playwright.async_api import Browser as PlaywrightBrowser
from playwright.async_api import Playwright, async_playwright
from browser_use.browser.context import BrowserContext, BrowserContextConfig, BrowserContextWindowSize
import os
import sys
from pathlib import Path

logger = logging.getLogger(__name__)


    
@dataclass
class BrowserConfig:
    """
    Configuration for the Browser.

    Default values:
            headless: False
                    Whether to run browser in headless mode

            disable_security: False
                    Disable browser security features

            extra_chromium_args: []
                    Extra arguments to pass to the browser

            wss_url: None
                    Connect to a browser instance via WebSocket

            cdp_url: None
                    Connect to a browser instance via CDP

            chrome_instance_path: None
                    Path to a Chrome instance to use to connect to your normal browser
                    e.g. '/Applications/Google\ Chrome.app/Contents/MacOS/Google\ Chrome'
                    
            user_data_dir: None
                    Path to a user data directory to enable persistent login
                    e.g. './user_data'
    """
    headless: bool = False
    disable_security: bool = True
    extra_chromium_args: list[str] = field(default_factory=list)
    browser_window_size:  BrowserContextWindowSize = field(default_factory=lambda: {
        'width': 1280 if sys.platform == 'linux' else 1600, 
        'height': 1024 if sys.platform == 'linux' else 900
    })
    chrome_instance_path: str | None = None
    wss_url: str | None = None
    cdp_url: str | None = None
    proxy: ProxySettings | None = field(default=None)
    user_data_dir: str | None = None
    new_context_config: BrowserContextConfig = field(default_factory=BrowserContextConfig)


class Browser:
    """
    Playwright browser on steroids.

    This is persistant browser factory that can spawn multiple browser contexts.
    It is recommended to use only one instance of Browser per your application (RAM usage will grow otherwise).
    """
    
    def __init__(self, config: BrowserConfig):
        """Initializing new browser"""
        logger.debug('Initializing new browser')
        self.config = config
        self.playwright = None
        self.playwright_browser = None
        self.disable_security_args = []
        if self.config.disable_security:
            self.disable_security_args = [
                '--disable-web-security',
                '--disable-site-isolation-trials',
                '--disable-features=IsolateOrigins,site-per-process'
            ]
    
    async def new_context(self, config: BrowserContextConfig):
        """Create a browser context"""
        return BrowserContext(config=config, browser=self)
    
    async def get_playwright_browser(self):
        """Get a browser context"""
        if not self.playwright_browser:
            await self._init()
        return self.playwright_browser
    
    async def _init(self):
        """Initialize the browser session"""
        self.playwright = await async_playwright().start()
        self.playwright_browser = await self._setup_browser(self.playwright)
        return self.playwright_browser
    
    async def _setup_cdp(self, playwright: Playwright):
        """Sets up and returns a Playwright Browser instance with anti-detection measures."""
        if not self.config.cdp_url:
            raise ValueError('CDP URL is required')
        
        logger.info(f'Connecting to remote browser via CDP {self.config.cdp_url}')
        return await playwright.chromium.connect_over_cdp(self.config.cdp_url)
    
    async def _setup_wss(self, playwright: Playwright):
        """Sets up and returns a Playwright Browser instance with anti-detection measures."""
        if not self.config.wss_url:
            raise ValueError('WSS URL is required')
        
        logger.info(f'Connecting to remote browser via WSS {self.config.wss_url}')
        return await playwright.chromium.connect(self.config.wss_url)
    
    async def _setup_browser_with_instance(self, playwright: Playwright):
        """Sets up and returns a Playwright Browser instance with anti-detection measures."""
        if not self.config.chrome_instance_path:
            raise ValueError('Chrome instance path is required')
        
        import requests
        timeout_seconds = 60
        
        logger.info('Try reusing existing Chrome instance')
        
        while timeout_seconds > 0:
            try:
                response = requests.get('http://localhost:9222/json/version', timeout=2)
                if response.status_code == 200:
                    return await playwright.chromium.connect_over_cdp(
                        endpoint_url='http://localhost:9222',
                        timeout=3000
                    )
            except requests.ConnectionError:
                logger.debug('No existing Chrome instance found, waiting for new one to start')
            
            await asyncio.sleep(1)
            timeout_seconds -= 1
        
        raise RuntimeError('Failed to connect to Chrome instance on localhost:9222')
    
    async def _setup_standard_browser(self, playwright: Playwright):
        """Sets up and returns a Playwright Browser instance with anti-detection measures."""
        import requests
        try:
            response = requests.get('http://localhost:9222/json/version', timeout=2)
            logger.info(f'Response: {response.json()}')
            
            if response.status_code == 200:
                logger.info(f'Connected to Chrome instance on localhost:9222 {response.json()["webSocketDebuggerUrl"]}')
                return await playwright.chromium.connect_over_cdp(endpoint_url=response.json()["webSocketDebuggerUrl"])
        except (requests.ConnectionError, requests.Timeout) as e:
            logger.debug(f'Failed to connect to existing Chrome instance: {e}')
            # 继续执行以启动新的浏览器实例
        
        user_data_dir = await self.create_user_data_dir(self.config)
        # 自动加载exts目录下所有扩展
        extensions_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '../../exts'))
        if os.path.exists(extensions_dir) and os.listdir(extensions_dir):
            ext_paths = [os.path.join(extensions_dir, d) for d in os.listdir(extensions_dir) if os.path.isdir(os.path.join(extensions_dir, d))]
            ext_paths_str = ','.join(ext_paths)
            extension_args = [
                f'--disable-extensions-except={ext_paths_str}',
                f'--load-extension={ext_paths_str}'
            ]
        else:
            extension_args = []
        
        await playwright.chromium.launch_persistent_context(
            user_data_dir=user_data_dir,
            headless=self.config.headless,
            viewport=self.config.browser_window_size,
            no_viewport=False,
            # user_agent=self.config.user_agent,
            device_scale_factor=2,
            java_script_enabled=True,
            # bypass_csp=self.config.disable_security,
            # ignore_https_errors=self.config.disable_security,
            # record_video_dir=self.config.save_recording_path,
            # record_video_size=self.config.browser_window_size,
            # locale=self.config.locale,
            args=[
                    '--no-sandbox',
                    '--disable-blink-features=AutomationControlled',
                    '--disable-infobars',
                    '--disable-background-timer-throttling',
                    '--disable-popup-blocking',
                    '--disable-backgrounding-occluded-windows',
                    '--disable-renderer-backgrounding',
                    '--disable-window-activation',
                    '--disable-focus-on-load',
                    '--disable-automation',
                    '--no-first-run',
                    '--no-default-browser-check',
                    "--disable-blink-features=AutomationControlled",
                    "--window-position=0,0 ",
                    "--enable-automation",
                    "--remote-debugging-port=9222"
                    # '--no-startup-window',
                    
                    # '--window-position=0,0'
                ] + self.disable_security_args + extension_args + self.config.extra_chromium_args,
                proxy=self.config.proxy
            )
    
        try:
            return await playwright.chromium.connect_over_cdp(
                endpoint_url='http://localhost:9222',
                timeout=3000
            )
        except Exception as e:
            logger.error(f'Failed to connect to launched Chrome instance: {e}')
            raise
    
    async def _setup_browser(self, playwright: Playwright):
        """Sets up and returns a Playwright Browser instance with anti-detection measures."""
        try:
            if self.config.cdp_url:
                return await self._setup_cdp(playwright)
            
            if self.config.wss_url:
                return await self._setup_wss(playwright)
            
            if self.config.chrome_instance_path:
                return await self._setup_browser_with_instance(playwright)
            
            return await self._setup_standard_browser(playwright)
        
        except Exception as e:
            logger.error(f'Failed to initialize Playwright browser: {str(e)}')
            raise
    
    async def close(self):
        """Close the browser instance"""
        try:
            logger.info('Closing browser instance')
            self.playwright_browser = None
            self.playwright = None
        
        except Exception as e:
            logger.debug(f'Failed to close browser properly: {e}')
        
        finally:
            self.playwright_browser = None
            self.playwright = None
    
    def __del__(self):
        """Async cleanup when object is destroyed"""
        if self.playwright_browser or self.playwright:
            try:
                loop = asyncio.get_running_loop()
                if loop.is_running():
                    loop.create_task(self.close())
                else:
                    asyncio.run(self.close())
            except Exception as e:
                logger.debug(f'Failed to cleanup browser in destructor: {e}')

    async def create_user_data_dir(self, config):
        if sys.platform.startswith('linux'):
            cache_dir = '/home/ubuntu/chrome_test/.cache'
            print(f'cache_dir: {cache_dir}')
        elif sys.platform == 'darwin':
            cache_dir = os.path.join(Path.home(), 'Library', 'Caches')
            print(f'cache_dir: {cache_dir}')
        elif sys.platform == 'win32':
            cache_dir = os.environ.get('LOCALAPPDATA', os.path.join(Path.home(), 'AppData', 'Local'))
            print(f'cache_dir: {cache_dir}')
        else:
            raise RuntimeError(f'Unsupported platform: {sys.platform}')
        channel = getattr(config, 'channel', None) or getattr(config, 'browserName', 'chromium')
        result = os.path.join(cache_dir, 'orion-playwright', f'{channel}')
        os.makedirs(result, exist_ok=True)
        return result


import logging
import time
from functools import wraps
from typing import Any, Callable, Coroutine, ParamSpec, TypeVar

logger = logging.getLogger(__name__)


# Define generic type variables for return type and parameters
R = TypeVar('R')
P = ParamSpec('P')


def time_execution_sync(additional_text: str = '') -> Callable[[Callable[P, R]], Callable[P, R]]:
	def decorator(func: Callable[P, R]) -> Callable[P, R]:
		@wraps(func)
		def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
			start_time = time.time()
			result = func(*args, **kwargs)
			execution_time = time.time() - start_time
			logger.debug(f'{additional_text} Execution time: {execution_time:.2f} seconds')
			return result

		return wrapper

	return decorator


def time_execution_async(
	additional_text: str = '',
) -> Callable[[Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]]:
	def decorator(func: Callable[P, Coroutine[Any, Any, R]]) -> Callable[P, Coroutine[Any, Any, R]]:
		@wraps(func)
		async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
			start_time = time.time()
			result = await func(*args, **kwargs)
			execution_time = time.time() - start_time
			logger.debug(f'{additional_text} Execution time: {execution_time:.2f} seconds')
			return result

		return wrapper

	return decorator


def singleton(cls):
	instance = [None]

	def wrapper(*args, **kwargs):
		if instance[0] is None:
			instance[0] = cls(*args, **kwargs)
		return instance[0]

	return wrapper



from setuptools import setup, find_packages
import os

# 读取requirements.txt中的依赖项
with open('requirements.txt') as f:
    requirements = f.read().splitlines()

# 读取README.md作为长描述
with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="orion_browser",
    version="0.4.5",
    author="BTY Team",
    author_email="qianhai@bantouyan.com",
    description="浏览器代理服务器",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bantouyan/orion",
    packages=find_packages(include=['orion_browser', 'orion_browser.*', 'app', 'app.*', 'app_data', 'browser_use', 'browser_use.*']),
    include_package_data=True,
    package_data={
        'app_data': ['js/*.js'],
        'browser_use': ['dom/*', 'browser/*', 'controller/*', 'agent/*', 'telemetry/*'],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'orion-server=app:start_server',
        ],
    },
) 


import logging
import os
import uuid
from pathlib import Path
from dotenv import load_dotenv
from posthog import Posthog
from browser_use.telemetry.views import BaseTelemetryEvent
from browser_use.utils import singleton

load_dotenv()
logger = logging.getLogger(__name__)

POSTHOG_EVENT_SETTINGS = {
    'process_person_profile': True
}

@singleton
class ProductTelemetry:
    """
    Service for capturing anonymized telemetry data.

    If the environment variable `ANONYMIZED_TELEMETRY=False`, anonymized telemetry will be disabled.
    """
    
    USER_ID_PATH = str(Path.home() / '.cache' / 'browser_use' / 'telemetry_user_id')
    PROJECT_API_KEY = 'phc_[REDACTED]' # API key was seen in bytecode
    HOST = 'https://eu.i.posthog.com'
    UNKNOWN_USER_ID = 'UNKNOWN'
    
    def __init__(self):
        # Check if telemetry is disabled via environment variable
        telemetry_disabled = os.getenv('ANONYMIZED_TELEMETRY', 'true').lower() == 'false'
        
        # Set debug logging based on environment variable
        self.debug_logging = os.getenv('BROWSER_USE_LOGGING_LEVEL', 'info').lower() == 'debug'
        
        if telemetry_disabled:
            self._posthog_client = None
        else:
            logging.info('Anonymized telemetry enabled. See https://docs.browser-use.com/development/telemetry for more information.')
            self._posthog_client = Posthog(
                project_api_key=self.PROJECT_API_KEY,
                host=self.HOST,
                disable_geoip=False
            )
            
            # Disable PostHog's own logging if we're not in debug mode
            if not self.debug_logging:
                posthog_logger = logging.getLogger('posthog')
                posthog_logger.disabled = True
        
        # Log if telemetry is disabled
        if self._posthog_client is None:
            logger.debug('Telemetry disabled')
        
        # Initialize user ID 
        self._curr_user_id = None
    
    def capture(self, event: BaseTelemetryEvent) -> None:
        """
        Capture a telemetry event
        """
        if self._posthog_client is None:
            return
        
        if self.debug_logging:
            logger.debug(f'Telemetry event: {event.name} {event.properties}')
        
        self._direct_capture(event)
    
    def _direct_capture(self, event: BaseTelemetryEvent) -> None:
        """
        Should not be thread blocking because posthog magically handles it
        """
        if self._posthog_client is None:
            return
        
        try:
            self._posthog_client.capture(
                self.user_id,
                event.name,
                {**event.properties, **POSTHOG_EVENT_SETTINGS}
            )
        except Exception as e:
            logger.error(f'Failed to send telemetry event {event.name}: {e}')
    
    @property
    def user_id(self) -> str:
        """
        Get or create a unique user ID for telemetry
        """
        if self._curr_user_id:
            return self._curr_user_id
            
        try:
            # Check if user ID file exists
            if not os.path.exists(self.USER_ID_PATH):
                # Create directory if it doesn't exist
                os.makedirs(os.path.dirname(self.USER_ID_PATH), exist_ok=True)
                
                # Generate and save a new user ID
                with open(self.USER_ID_PATH, 'w') as f:
                    user_id = str(uuid.uuid4())
                    f.write(user_id)
                
                self._curr_user_id = user_id
            else:
                # Read existing user ID
                with open(self.USER_ID_PATH, 'r') as f:
                    self._curr_user_id = f.read()
        except Exception:
            # Use fallback user ID if any error occurs
            self._curr_user_id = self.UNKNOWN_USER_ID
            
        return self._curr_user_id